"""
Tests for the Talos SDK MCP module.
"""

from talos_sdk import Wallet, sign_mcp_request, verify_mcp_response


class TestMCPSigning:
    """Tests for MCP signing functions."""

    def test_sign_mcp_request_creates_frame(self):
        """sign_mcp_request should create a valid SignedFrame."""
        wallet = Wallet.generate()
        request = {"method": "read", "params": {"path": "/data"}}

        frame = sign_mcp_request(
            wallet=wallet,
            request=request,
            session_id="session-123",
            correlation_id="corr-456",
            tool="filesystem",
            action="read",
            timestamp=1767500000,
        )

        assert frame.correlation_id == "corr-456"
        assert frame.signer_did == wallet.to_did()
        assert len(frame.signature) == 64
        assert len(frame.payload) > 0

    def test_sign_mcp_is_deterministic(self):
        """Same inputs should produce same signature."""
        seed = bytes(32)
        wallet = Wallet.from_seed(seed)
        request = {"method": "test"}

        frame1 = sign_mcp_request(
            wallet, request, "s1", "c1", "tool", "action", timestamp=1000
        )
        frame2 = sign_mcp_request(
            wallet, request, "s1", "c1", "tool", "action", timestamp=1000
        )

        assert frame1.signature == frame2.signature
        assert frame1.payload == frame2.payload

    def test_verify_mcp_response_valid(self):
        """verify_mcp_response should pass for valid frame."""
        wallet = Wallet.generate()
        frame = sign_mcp_request(wallet, {"method": "test"}, "s", "c", "t", "a")

        is_valid = verify_mcp_response(frame, "c", wallet.public_key)
        assert is_valid is True

    def test_verify_mcp_response_wrong_correlation(self):
        """Wrong correlation ID should fail verification."""
        wallet = Wallet.generate()
        frame = sign_mcp_request(wallet, {"method": "test"}, "s", "correct", "t", "a")

        is_valid = verify_mcp_response(frame, "wrong", wallet.public_key)
        assert is_valid is False

    def test_verify_mcp_response_wrong_key(self):
        """Wrong public key should fail verification."""
        wallet1 = Wallet.generate()
        wallet2 = Wallet.generate()
        frame = sign_mcp_request(wallet1, {"method": "test"}, "s", "c", "t", "a")

        is_valid = verify_mcp_response(frame, "c", wallet2.public_key)
        assert is_valid is False



import abc
import binascii
import json
import re
from datetime import datetime, timezone
from typing import Optional, Dict, Any
from pydantic import BaseModel, Field, field_validator

ALGORITHM_AES_256_GCM = "aes-256-gcm"
SCHEMA_ID_ENVELOPE = "talos.secrets.envelope"
SCHEMA_VERSION_V1 = "v1"

class Envelope(BaseModel):
    """
    Encrypted data envelope (Draft 2020-12 / Normative).
    
    Ensures structural integrity and metadata compliance for secrets-at-rest.
    """
    kek_id: str = Field(..., min_length=1, max_length=255)
    iv: str = Field(..., pattern=r"^[0-9a-f]{24}$")        # 24 hex char (12 bytes)
    ciphertext: str = Field(..., pattern=r"^[0-9a-f]+$") # Hex
    tag: str = Field(..., pattern=r"^[0-9a-f]{32}$")       # 32 hex char (16 bytes)
    alg: str = Field(default=ALGORITHM_AES_256_GCM)
    schema_id: str = Field(default=SCHEMA_ID_ENVELOPE)
    schema_version: str = Field(default=SCHEMA_VERSION_V1)
    created_at: str = Field(default_factory=lambda: datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z")

    @field_validator("alg")
    @classmethod
    def validate_alg(cls, v):
        if v != ALGORITHM_AES_256_GCM:
            raise ValueError(f"Unsupported algorithm: {v}")
        return v

    @field_validator("schema_id")
    @classmethod
    def validate_schema_id(cls, v):
        if v != SCHEMA_ID_ENVELOPE:
            raise ValueError(f"Invalid schema_id: {v}")
        return v

    @field_validator("schema_version")
    @classmethod
    def validate_schema_version(cls, v):
        if v != SCHEMA_VERSION_V1:
            raise ValueError(f"Invalid schema_version: {v}")
        return v

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'Envelope':
        return Envelope.model_validate(data)

    def to_json(self) -> str:
        return self.model_dump_json()

    @staticmethod
    def from_json(json_str: str) -> 'Envelope':
        return Envelope.model_validate_json(json_str)

class KekProvider(abc.ABC):
    """Abstract interface for Key Encryption Key providers."""

    @abc.abstractmethod
    def encrypt(self, plaintext: bytes) -> Envelope:
        """Encrypt plaintext bytes and return Envelope."""

    @abc.abstractmethod
    def decrypt(self, envelope: Envelope) -> bytes:
        """Decrypt envelope and return plaintext bytes."""

def generate_master_key() -> str:
    """Generate a stable 256-bit (32 byte) master key in hex format."""
    import secrets
    return secrets.token_hex(32)



"""Talos SDK Canonical JSON.

Implements the canonicalization rules from CANONICAL_JSON.md.
"""

import json
from typing import Any


def canonical_json(obj: Any) -> str:
    """Serialize an object to canonical JSON.

    Rules:
    - Keys sorted lexicographically
    - No whitespace outside strings
    - UTF-8 encoding
    - Minimal escaping

    Args:
        obj: Object to serialize

    Returns:
        Canonical JSON string
    """

    def _preprocess(o: Any) -> Any:
        if isinstance(o, dict):
            return {k: _preprocess(v) for k, v in o.items()}
        if isinstance(o, list):
            return [_preprocess(v) for v in o]
        if isinstance(o, float) and o.is_integer():
            return int(o)
        return o

    return json.dumps(
        _preprocess(obj),
        sort_keys=True,
        separators=(",", ":"),
        ensure_ascii=False,
    )


def canonical_json_bytes(obj: Any) -> bytes:
    """Serialize an object to canonical JSON bytes.

    Args:
        obj: Object to serialize

    Returns:
        UTF-8 encoded canonical JSON bytes
    """
    return canonical_json(obj).encode("utf-8")



"""Sequence tracking for sender_seq monotonicity."""

import threading
from typing import Protocol


class SequenceStorage(Protocol):
    """Interface for persisting sequence state."""

    def load(self, session_id: str, sender_id: str) -> int | None:
        """Load the next sequence number for (session_id, sender_id)."""
        ...

    def save(self, session_id: str, sender_id: str, next_seq: int) -> None:
        """Save the next sequence number for (session_id, sender_id)."""
        ...


class InMemorySequenceStorage:
    """Thread-safe in-memory sequence storage."""

    def __init__(self):
        self._data: dict[tuple[str, str], int] = {}
        self._lock = threading.Lock()

    def load(self, session_id: str, sender_id: str) -> int | None:
        with self._lock:
            return self._data.get((session_id, sender_id))

    def save(self, session_id: str, sender_id: str, next_seq: int) -> None:
        with self._lock:
            self._data[(session_id, sender_id)] = next_seq


# Process-global singleton for consistent sequence tracking across instances
_DEFAULT_STORAGE = InMemorySequenceStorage()


class SequenceTracker:
    """Tracks sender_seq per (session_id, sender_id) with atomic reserve().

    Usage:
        tracker = SequenceTracker.load(session_id, sender_id)
        seq = tracker.reserve()  # Atomically get and persist next seq
    """

    def __init__(
        self,
        session_id: str,
        sender_id: str,
        storage: SequenceStorage | None = None,
    ):
        self._session_id = session_id
        self._sender_id = sender_id
        self._storage = storage or _DEFAULT_STORAGE
        self._next_seq: int = self._storage.load(session_id, sender_id) or 0
        self._lock = threading.Lock()

    def reserve(self) -> int:
        """Atomically reserve and persist the next sequence number.

        Returns:
            The reserved sequence number (0-indexed, monotonically increasing).
        """
        with self._lock:
            seq = self._next_seq
            self._next_seq += 1
            self._storage.save(self._session_id, self._sender_id, self._next_seq)
            return seq

    def current(self) -> int:
        """Get current next sequence without incrementing."""
        return self._next_seq

    @classmethod
    def load(
        cls, session_id: str, sender_id: str, storage: SequenceStorage | None = None
    ) -> "SequenceTracker":
        """Load or create a SequenceTracker for (session_id, sender_id)."""
        return cls(session_id, sender_id, storage)



"""A2A Error types and mapping factory."""

from typing import TYPE_CHECKING

from talos_sdk.errors import TalosError

if TYPE_CHECKING:
    from .models import ErrorResponse


class A2AError(TalosError):
    """Base class for A2A errors."""
    pass


class A2ATransportError(A2AError):
    """HTTP transport failure."""

    def __init__(self, message: str, status_code: int | None = None):
        super().__init__("A2A_TRANSPORT_ERROR", message)
        self.status_code = status_code


class A2ASessionNotFoundError(A2AError):
    """Session not found."""

    def __init__(self):
        super().__init__("A2A_SESSION_NOT_FOUND", "Session not found")


class A2ASessionStateInvalidError(A2AError):
    """Invalid session state transition."""

    def __init__(self, current_state: str = ""):
        super().__init__("A2A_SESSION_STATE_INVALID", f"Invalid state: {current_state}")
        self.current_state = current_state


class A2AFrameReplayError(A2AError):
    """Duplicate frame detected (replay)."""

    def __init__(self, sender_seq: int = -1):
        super().__init__("A2A_FRAME_REPLAY_DETECTED", f"Duplicate sender_seq: {sender_seq}")
        self.sender_seq = sender_seq


class A2AFrameDigestMismatchError(A2AError):
    """Frame digest does not match computed value."""

    def __init__(self):
        super().__init__("A2A_FRAME_DIGEST_MISMATCH", "Frame digest mismatch")


class A2AFrameSequenceTooFarError(A2AError):
    """Sequence number exceeds MAX_FUTURE_DELTA."""

    def __init__(self, sender_seq: int = -1, expected_seq: int = -1):
        super().__init__("A2A_FRAME_SEQUENCE_TOO_FAR", f"Sequence too far: {sender_seq}. Expected: {expected_seq}")
        self.sender_seq = sender_seq
        self.expected_seq = expected_seq


class A2AFrameSizeExceededError(A2AError):
    """Frame payload exceeds maximum size."""

    def __init__(self, size: int = -1, limit: int = -1):
        super().__init__("A2A_FRAME_SIZE_EXCEEDED", f"Frame size {size} exceeds limit {limit}")
        self.size = size
        self.limit = limit


class A2ACryptoNotConfiguredError(A2AError):
    """FrameCrypto hook not configured (Phase 10.3 required)."""

    def __init__(self):
        super().__init__("A2A_CRYPTO_NOT_CONFIGURED", "FrameCrypto not configured")


def raise_mapped_error(err: "ErrorResponse", status_code: int | None = None) -> None:
    """Map gateway ErrorResponse to SDK exception.

    Extracts details from err.details when available.
    """
    code = err.error_code
    details = err.details or {}

    if code == "A2A_SESSION_NOT_FOUND":
        raise A2ASessionNotFoundError()

    if code == "A2A_SESSION_STATE_INVALID":
        raise A2ASessionStateInvalidError(current_state=str(details.get("current_state", "")))

    if code == "A2A_FRAME_REPLAY_DETECTED":
        seq = details.get("sender_seq", -1)
        raise A2AFrameReplayError(sender_seq=int(seq) if seq is not None else -1)

    if code == "A2A_FRAME_DIGEST_MISMATCH":
        raise A2AFrameDigestMismatchError()

    if code == "A2A_FRAME_SEQUENCE_TOO_FAR":
        seq = details.get("sender_seq", details.get("received_seq", -1))
        expected = details.get("expected_seq", -1)
        raise A2AFrameSequenceTooFarError(
            sender_seq=int(seq) if seq is not None else -1,
            expected_seq=int(expected) if expected is not None else -1
        )

    if code == "A2A_FRAME_SIZE_EXCEEDED":
        size = details.get("size", -1)
        limit = details.get("limit", -1)
        raise A2AFrameSizeExceededError(
            size=int(size) if size is not None else -1,
            limit=int(limit) if limit is not None else -1
        )

    # Default fallback
    raise A2ATransportError(err.message, status_code=status_code)



import argparse
import sys

from .runner import run_conformance


def main() -> None:
    parser = argparse.ArgumentParser(description="Talos SDK Conformance Harness")
    parser.add_argument(
        "--vectors", required=True, help="Path to test vector JSON file"
    )
    parser.add_argument("--report", help="Path to write JUnit XML report")

    args = parser.parse_args()

    try:
        success = run_conformance(args.vectors, args.report)
        sys.exit(0 if success else 1)
    except Exception as e:
        print(f"Error handling vectors: {e}", file=sys.stderr)
        sys.exit(2)


if __name__ == "__main__":
    main()



import json

from talos_sdk.crypto import generate_encryption_keypair
from talos_sdk.session import Session, SessionManager
from talos_sdk.wallet import Wallet


def test_ratchet_flow_and_out_of_order():
    # 1. Setup Identities
    alice_wallet = Wallet.generate("Alice")
    bob_wallet = Wallet.generate("Bob")

    # NOTE: In this implementation SessionManager takes a KeyPair not a Wallet
    # We need to construct a KeyPair from the wallet's keys for SessionManager
    # Or SessionManager expects ED25519 helper?
    # Let's check SessionManager.__init__: def __init__(self, identity_keypair: KeyPair ...
    # And KeyPair is from .crypto.
    # Wallet uses Ed25519PrivateKey directly.
    # We need to adapt because SessionManager uses KeyPair struct.

    from talos_sdk.crypto import KeyPair

    def wallet_to_keypair(w):
        return KeyPair(
            public_key=w.public_key,
            private_key=w._private_key.private_bytes_raw(),
            key_type="ed25519",
        )

    alice_mgr = SessionManager(wallet_to_keypair(alice_wallet))
    bob_mgr = SessionManager(wallet_to_keypair(bob_wallet))

    # 2. Publish Bob's Bundle
    bob_bundle = bob_mgr.get_prekey_bundle()

    # 3. Alice initiates session
    alice_session = alice_mgr.create_session_as_initiator("Bob", bob_bundle)

    # 4. Alice sends message "Hello Bob"
    msg1 = alice_session.encrypt(b"Hello Bob")

    # 5. Bob receives "Hello Bob"
    # Bob needs to create session from first message hint?
    # Or does create_session_as_responder assume we know the peer?
    # In Signal X3DH, the first message contains the initial key exchange info usually.
    # Here create_session_as_responder signature is:
    # create_session_as_responder(self, peer_id: str, peer_dh_public: bytes)
    # The first message envelope contains "header": {"dh": ...}

    envelope1 = json.loads(msg1)
    # Need to base64url decode envelope1["header"]["dh"] to get bytes
    from talos_sdk.crypto import b64u_decode

    alice_dh_public = b64u_decode(envelope1["header"]["dh"])

    bob_session = bob_mgr.create_session_as_responder("Alice", alice_dh_public)
    decrypted1 = bob_session.decrypt(msg1)
    assert decrypted1 == b"Hello Bob"

    # 6. Bob replies "Hello Alice"
    msg2 = bob_session.encrypt(b"Hello Alice")
    decrypted2 = alice_session.decrypt(msg2)
    assert decrypted2 == b"Hello Alice"

    # 7. Ratchet Test (Ping Pong)
    msgs = [b"MSG_1", b"MSG_2", b"MSG_3"]
    encrypted_msgs = []

    for m in msgs:
        encrypted_msgs.append(alice_session.encrypt(m))

    # Decrypt out of order
    # Decrypt 3 first, then 1, then 2
    assert bob_session.decrypt(encrypted_msgs[2]) == msgs[2]
    assert bob_session.decrypt(encrypted_msgs[0]) == msgs[0]
    assert bob_session.decrypt(encrypted_msgs[1]) == msgs[1]


def test_session_serialization():
    # Setup
    # Setup
    # w = Wallet.generate("Test") - REMOVED (Unused)
    # kp = KeyPair(...) - REMOVED (Unused)
    # SessionManager not needed for this test; we are testing Session serialization directly.

    # Create a dummy session (mocked state for minimal test)
    # Ideally should run thorough handshake, but re-using flow logic above
    # Let's just create a session and serialize it

    # Manually bootstrapping state just to test ser/deser
    from talos_sdk.session import RatchetState

    state = RatchetState(
        dh_keypair=generate_encryption_keypair(),
        dh_remote=generate_encryption_keypair().public_key,
        root_key=b"0" * 32,
        chain_key_send=b"1" * 32,
        chain_key_recv=b"2" * 32,
    )

    sess = Session("remote-peer", state)
    data = sess.to_dict()

    sess2 = Session.from_dict(data)
    assert sess2.peer_id == sess.peer_id
    assert sess2.state.root_key == sess.state.root_key



from abc import ABC, abstractmethod


class IKeyValueStorePort(ABC):
    """Port for Key-Value storage operations.
    Used for projection state, caching, and metadata.
    """

    @abstractmethod
    def get(self, key: str) -> bytes | None:
        """Retrieve a value by key. Returns None if not found."""
        ...

    @abstractmethod
    def put(self, key: str, value: bytes) -> None:
        """Store a value by key. Overwrites existing value."""
        ...

    @abstractmethod
    def delete(self, key: str) -> None:
        """Delete a key-value pair."""
        ...



import json
import os
import pytest
from talos_sdk.validation import validate_identity, IdentityValidationError

# Path to vectors
# Path to vectors
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../"))
VECTORS_PATH = os.path.join(ROOT_DIR, "contracts/test_vectors/sdk/identity_vectors.json")

def load_matrix():
    with open(VECTORS_PATH, "r") as f:
        return json.load(f)

MATRIX = load_matrix()

def get_test_cases():
    cases = []
    for schema_type, categories in MATRIX.items():
        for category, vectors in categories.items():
            is_valid = (category == "valid")
            for v in vectors:
                cases.append((schema_type, v["data"], is_valid, v["name"]))
    return cases

@pytest.mark.parametrize("schema_type, data, expected_valid, case_name", get_test_cases())
def test_identity_conformance(schema_type, data, expected_valid, case_name):
    """Verify that SDK validation matches normative contract vectors."""
    
    if expected_valid:
        # Should not raise
        validate_identity(data, schema_type)
    else:
        # Should raise IdentityValidationError
        with pytest.raises(IdentityValidationError):
            validate_identity(data, schema_type)

def test_missing_schema():
    with pytest.raises(IdentityValidationError, match="Schema for missing not found"):
        validate_identity({}, "missing")

def test_invalid_id_casing():
    # Explicit check for lowercase uuidv7 as per spec
    bad_principal = {
        "schema_id": "talos.principal",
        "schema_version": "v2",
        "id": "01945533-3158-7C85-992D-9865F1715698", # UPPERCASE FAIL
        "principal_id": "p-1",
        "type": "user",
        "auth_mode": "bearer",
        "status": "active",
        "team_id": "01945533-3158-7c85-992d-9865f1715699"
    }
    with pytest.raises(IdentityValidationError, match="Validation failed"):
        validate_identity(bad_principal, "principal")

def test_extra_property_rejection():
    # proving additionalProperties: false
    leaky_org = {
        "schema_id": "talos.org",
        "schema_version": "v2",
        "id": "01945533-3158-7c85-992d-9865f1715698",
        "name": "Acme",
        "domain": "acme.com",
        "extra": "leak"
    }
    with pytest.raises(IdentityValidationError, match="Validation failed"):
        validate_identity(leaky_org, "org")



"""Capability management for Talos SDK."""

import base64
from typing import Any

from .canonical import canonical_json_bytes
from .wallet import Wallet


def base64url_encode(b: bytes) -> str:
    return base64.urlsafe_b64encode(b).decode("utf-8").rstrip("=")


def base64url_decode(s: str) -> bytes:
    s += "=" * ((4 - len(s) % 4) % 4)
    return base64.urlsafe_b64decode(s)


class Capability:
    def __init__(self, data: dict[str, Any]):
        self.data = data
        self.v = data.get("v", "1")
        self.iss = data.get("iss")
        self.sub = data.get("sub")
        self.scope = data.get("scope")
        self.iat = data.get("iat")
        self.exp = data.get("exp")
        self.sig = data.get("sig")

    @classmethod
    def create(
        cls,
        issuer_wallet: Wallet,
        subject_did: str,
        scope: Any,
        exp: int,
        iat: int | None = None,
    ) -> "Capability":
        if iat is None:
            import time

            iat = int(time.time())

        cap_data = {
            "v": "1",
            "iss": issuer_wallet.to_did(),
            "sub": subject_did,
            "scope": scope,
            "iat": iat,
            "exp": exp,
        }

        canon = canonical_json_bytes(cap_data)
        sig = issuer_wallet.sign(canon)
        cap_data["sig"] = base64url_encode(sig)

        return cls(cap_data)

    def verify(self, issuer_public_key: bytes, now: int | None = None) -> bool:
        if not self.sig:
            return False

        # Verify expiry
        if now is None:
            import time

            now = int(time.time())

        if self.exp is None or self.exp < now:
            return False

        # Get content without signature
        content = {k: v for k, v in self.data.items() if k != "sig"}
        canon = canonical_json_bytes(content)
        sig_bytes = base64url_decode(self.sig)

        return Wallet.verify(canon, sig_bytes, issuer_public_key)

    def authorize(self, tool: str, action: str) -> bool:
        """Simple scope check: scope is list of {tool, actions}"""
        if not isinstance(self.scope, list):
            return False

        for s in self.scope:
            if s.get("tool") == tool and action in s.get("actions", []):
                return True
        return False



from abc import ABC, abstractmethod
from typing import Protocol

# Placeholder for models - in a real scenario these would come from talos-contracts
# But since talos-contracts might not be in the python path of this agent environment,
# I will define Protocol or use Any for now, or assume imports work.
# The user's prompt showed `event: AuditEvent`.
# I will try to import from talos_contracts if possible, or define type aliases.


class AuditEvent(Protocol):
    event_id: str
    timestamp: float
    # ...


class Filters(Protocol):
    pass


class EventPage(Protocol):
    events: list[AuditEvent]
    next_cursor: str | None


class Stats(Protocol):
    count: int


class TimeWindow(Protocol):
    start: float
    end: float


class IAuditStorePort(ABC):
    """Port for storing and retrieving audit events.
    Cohesive grouping of audit log storage operations.
    """

    @abstractmethod
    def append(self, event: AuditEvent) -> None:
        """Append a new audit event to the immutable log."""
        ...

    @abstractmethod
    def list(
        self,
        before: str | None = None,
        limit: int = 100,
        filters: Filters | None = None,
    ) -> EventPage:
        """List events with pagination (cursor-based) and filtering."""
        ...

    @abstractmethod
    def stats(self, window: TimeWindow) -> Stats:
        """Retrieve statistics for a time window."""
        ...



import os
import requests
from typing import Dict, Any, List, Optional

class McpClient:
    def __init__(self, gateway_url: str, api_token: Optional[str] = None):
        self.gateway_url = gateway_url.rstrip("/")
        self.api_token = api_token or os.getenv("TALOS_API_TOKEN")

    def _get_headers(self) -> Dict[str, str]:
        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers["Authorization"] = f"Bearer {self.api_token}"
        return headers

    def list_servers(self) -> List[Dict[str, Any]]:
        """List available MCP servers known to the gateway."""
        url = f"{self.gateway_url}/v1/mcp/servers"
        resp = requests.get(url, headers=self._get_headers(), timeout=10)
        resp.raise_for_status()
        return resp.json().get("servers", [])

    def list_tools(self, server_id: str) -> List[Dict[str, Any]]:
        """List tools provided by a specific server."""
        url = f"{self.gateway_url}/v1/mcp/servers/{server_id}/tools"
        resp = requests.get(url, headers=self._get_headers(), timeout=10)
        resp.raise_for_status()
        return resp.json().get("tools", [])

    def get_tool_schema(self, server_id: str, tool_name: str) -> Dict[str, Any]:
        """Get the JSON schema for a tool."""
        url = f"{self.gateway_url}/v1/mcp/servers/{server_id}/tools/{tool_name}/schema"
        resp = requests.get(url, headers=self._get_headers(), timeout=10)
        resp.raise_for_status()
        return resp.json().get("json_schema", {})

    def invoke_tool(self, server_id: str, tool_name: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """Invoke a tool via the gateway."""
        url = f"{self.gateway_url}/v1/mcp/servers/{server_id}/tools/{tool_name}:call"
        payload = {"input": arguments}
        
        resp = requests.post(url, json=payload, headers=self._get_headers(), timeout=30)
        resp.raise_for_status()
        
        data = resp.json()
        if "error" in data:
            raise RuntimeError(f"MCP Invocation Error: {data['error']}")
            
        return data.get("output", {})



"""Talos SDK Wallet.

Identity management as defined in SDK_CONTRACT.md.
"""

import hashlib

from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)

from .errors import TalosInvalidInputError

# Base58btc alphabet for DID encoding
_BASE58_ALPHABET = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def _base58_encode(data: bytes) -> str:
    """Encode bytes to base58btc."""
    num = int.from_bytes(data, "big")
    result = []
    while num > 0:
        num, remainder = divmod(num, 58)
        result.append(_BASE58_ALPHABET[remainder])
    # Handle leading zeros
    for byte in data:
        if byte == 0:
            result.append(_BASE58_ALPHABET[0])
        else:
            break
    return "".join(reversed(result))


class Wallet:
    """Talos identity wallet for key management and signing.

    Implements the Identity module from SDK_CONTRACT.md.
    """

    def __init__(
        self,
        private_key: Ed25519PrivateKey,
        name: str | None = None,
    ):
        """Initialize a Wallet from a private key.

        Args:
            private_key: Ed25519 private key
            name: Optional human-readable name
        """
        self._private_key = private_key
        self._public_key = private_key.public_key()
        self._name = name

    @classmethod
    def generate(cls, name: str | None = None) -> "Wallet":
        """Generate a new wallet with a random keypair.

        Args:
            name: Optional human-readable name for the wallet

        Returns:
            A new Wallet instance with a randomly generated keypair

        Note:
            This method is non-deterministic (uses secure random).
        """
        private_key = Ed25519PrivateKey.generate()
        return cls(private_key, name)

    @classmethod
    def from_seed(cls, seed: bytes, name: str | None = None) -> "Wallet":
        """Create a wallet from a 32-byte seed.

        Args:
            seed: 32-byte seed (raw bytes)
            name: Optional human-readable name

        Returns:
            A Wallet instance derived deterministically from the seed

        Raises:
            TalosInvalidInputError: If seed length is not 32 bytes
        """
        if len(seed) != 32:
            raise TalosInvalidInputError(
                f"Seed must be exactly 32 bytes, got {len(seed)}",
                details={"seed_length": len(seed)},
            )
        private_key = Ed25519PrivateKey.from_private_bytes(seed)
        return cls(private_key, name)

    def to_did(self) -> str:
        """Convert the wallet's public key to a DID string.

        Returns:
            DID string in format did:key:z6Mk...

        Note:
            Deterministic - same public key produces identical DID.
        """
        # Multicodec prefix for Ed25519 public key: 0xed01
        prefix = bytes([0xED, 0x01])
        public_bytes = self._public_key.public_bytes_raw()
        multicodec_key = prefix + public_bytes
        return f"did:key:z{_base58_encode(multicodec_key)}"

    @property
    def address(self) -> str:
        """Get the hex-encoded public key hash (address).

        Returns:
            64-character hex string
        """
        public_bytes = self._public_key.public_bytes_raw()
        return hashlib.sha256(public_bytes).hexdigest()

    @property
    def public_key(self) -> bytes:
        """Get the 32-byte public key."""
        return self._public_key.public_bytes_raw()

    @property
    def name(self) -> str | None:
        """Get the wallet name."""
        return self._name

    @property
    def key_id(self) -> str:
        """Get the key ID (truncated DID for HTTP headers)."""
        return self.to_did()

    def sign(self, message: bytes) -> bytes:
        """Sign a message using Ed25519.

        Args:
            message: Arbitrary message bytes

        Returns:
            64-byte Ed25519 signature

        Note:
            Deterministic - Ed25519 signing is deterministic.
        """
        return self._private_key.sign(message)

    @staticmethod
    def verify(message: bytes, signature: bytes, public_key: bytes) -> bool:
        """Verify a signature against a message and public key.

        Args:
            message: Original message bytes
            signature: 64-byte signature
            public_key: 32-byte public key

        Returns:
            True if signature is valid, False otherwise
        """
        try:
            pk = Ed25519PublicKey.from_public_bytes(public_key)
            pk.verify(signature, message)
            return True
        except Exception:
            return False



"""RatchetFrameCrypto - bridges A2A transport with Double Ratchet Session.

Phase 10.3 implementation per LOCKED SPEC:
- Nonce fixed at 12 bytes (ChaCha20-Poly1305)
- ciphertext_hash is SHA-256 of ciphertext only (excluding nonce)
- frame_digest computed externally by session_client
"""

import base64
import hashlib
import json
from typing import TYPE_CHECKING

from talos_sdk.canonical import canonical_json_bytes
from talos_sdk.crypto import b64u_encode

if TYPE_CHECKING:
    from talos_sdk.session import Session

# ChaCha20-Poly1305 nonce length (fixed, hardcoded)
NONCE_LEN = 12


def _b64u_decode_strict(s: str) -> bytes:
    """Decode base64url without padding.

    Rejects:
    - Inputs containing '='
    - Impossible lengths (len % 4 == 1)

    Raises:
        ValueError: If input is invalid base64url
    """
    if "=" in s:
        raise ValueError("Base64url must not contain padding")

    # base64url without padding can never have length mod 4 == 1
    rem = len(s) % 4
    if rem == 1:
        raise ValueError("Invalid base64url length")

    # Add internal padding only for decoding
    if rem:
        s = s + ("=" * (4 - rem))

    return base64.urlsafe_b64decode(s)


class RatchetFrameCrypto:
    """FrameCrypto implementation using Double Ratchet Session.

    Bridges Phase 10.2 transport with existing ratchet encryption.
    Does NOT compute frame_digest (caller computes after reserving sender_seq).

    Usage:
        crypto = RatchetFrameCrypto(session)
        header_b64u, ciphertext_b64u, ciphertext_hash = crypto.encrypt(plaintext)
    """

    def __init__(self, session: "Session"):
        self._session = session

    def encrypt(self, plaintext: bytes) -> tuple[str, str, str]:
        """Encrypt plaintext and return frame components.

        Returns:
            Tuple of (header_b64u, ciphertext_b64u, ciphertext_hash)

        Note:
            frame_digest is computed by session_client after reserving sender_seq.

        Raises:
            ValueError: If nonce length is unexpected (regression guard)
        """
        # Use existing session encrypt
        envelope_bytes = self._session.encrypt(plaintext)
        envelope = json.loads(envelope_bytes.decode("utf-8"))

        # Extract and validate components
        header_dict = envelope["header"]
        nonce = _b64u_decode_strict(envelope["nonce"])
        ciphertext = _b64u_decode_strict(envelope["ciphertext"])

        # Regression guard: assert nonce length
        if len(nonce) != NONCE_LEN:
            raise ValueError(f"Unexpected nonce length: {len(nonce)}, expected {NONCE_LEN}")

        # Compute header_b64u from canonical header
        # Uses same format Session expects: {"dh", "n", "pn"}
        header_bytes = canonical_json_bytes(header_dict)
        header_b64u = b64u_encode(header_bytes)

        # Combine nonce + ciphertext for transport
        combined = nonce + ciphertext
        ciphertext_b64u = b64u_encode(combined)

        # Compute ciphertext_hash (SHA256 of raw ciphertext only, not nonce)
        ciphertext_hash = hashlib.sha256(ciphertext).hexdigest()

        return header_b64u, ciphertext_b64u, ciphertext_hash

    def decrypt(
        self,
        header_b64u: str,
        ciphertext_b64u: str,
        ciphertext_hash: str | None = None,
    ) -> bytes:
        """Decrypt frame and return plaintext.

        Args:
            header_b64u: Base64url encoded canonical header JSON
            ciphertext_b64u: Base64url encoded nonce||ciphertext
            ciphertext_hash: Optional SHA256 hash for integrity check

        Returns:
            Decrypted plaintext bytes

        Raises:
            ValueError: If base64url is invalid, ciphertext too short,
                       or ciphertext_hash mismatch
            RatchetError: If decryption fails (from Session.decrypt)
        """
        # Decode components (strict: reject padding and invalid lengths)
        header_bytes = _b64u_decode_strict(header_b64u)
        combined = _b64u_decode_strict(ciphertext_b64u)

        # Split nonce (12 bytes) and ciphertext
        if len(combined) < NONCE_LEN:
            raise ValueError(f"Ciphertext too short: {len(combined)} bytes")
        nonce = combined[:NONCE_LEN]
        ciphertext = combined[NONCE_LEN:]

        # Regression guard: assert nonce length after split
        if len(nonce) != NONCE_LEN:
            raise ValueError(f"Unexpected nonce length after split: {len(nonce)}")

        # Optional integrity check
        if ciphertext_hash is not None:
            computed_hash = hashlib.sha256(ciphertext).hexdigest()
            if computed_hash != ciphertext_hash:
                raise ValueError(
                    f"Ciphertext hash mismatch: expected {ciphertext_hash}, got {computed_hash}"
                )

        # Parse header JSON safely
        try:
            header_dict = json.loads(header_bytes.decode("utf-8"))
        except (UnicodeDecodeError, json.JSONDecodeError) as e:
            raise ValueError(f"Invalid header JSON: {e}") from e

        # Validate required header fields (ratchet expects these)
        required_fields = {"dh", "pn", "n"}
        if not required_fields.issubset(header_dict.keys()):
            missing = required_fields - set(header_dict.keys())
            raise ValueError(f"Missing header fields: {missing}")

        # Reconstruct envelope in the format Session.decrypt expects
        envelope = {
            "header": header_dict,
            "nonce": b64u_encode(nonce),
            "ciphertext": b64u_encode(ciphertext),
        }

        # Serialize using canonical_json_bytes (same as Session.encrypt output)
        envelope_bytes = canonical_json_bytes(envelope)

        return self._session.decrypt(envelope_bytes)



import pytest
import binascii
import os
import json
from talos_sdk.secrets.kek import Envelope, ALGORITHM_AES_256_GCM, generate_master_key, SCHEMA_ID_ENVELOPE
from talos_sdk.secrets.local_provider import LocalKekProvider

# Do NOT hardcode master keys in source. Load from environment or generate for ephemeral tests.
MASTER_KEY = os.environ.get("TALOS_MASTER_KEY", generate_master_key())
KEY_ID = "kek-v1"

@pytest.fixture
def provider():
    return LocalKekProvider(MASTER_KEY, KEY_ID)

def test_encrypt_decrypt_roundtrip(provider):
    plaintext = b"super secret data"
    envelope = provider.encrypt(plaintext)
    
    # Verify Envelope Structure
    assert envelope.kek_id == KEY_ID
    assert envelope.alg == ALGORITHM_AES_256_GCM
    assert envelope.schema_id == SCHEMA_ID_ENVELOPE
    assert len(binascii.unhexlify(envelope.iv)) == 12
    assert len(binascii.unhexlify(envelope.tag)) == 16
    assert envelope.ciphertext # Not empty
    assert envelope.created_at.endswith("Z")
    
    # Decrypt
    decrypted = provider.decrypt(envelope)
    assert decrypted == plaintext

def test_decrypt_fail_key_mismatch(provider):
    plaintext = b"data"
    envelope = provider.encrypt(plaintext)
    
    # Create new envelope with different key ID but same data
    data = envelope.to_dict()
    data["kek_id"] = "wrong-key"
    bad_envelope = Envelope.from_dict(data)
    
    with pytest.raises(ValueError, match="Key mismatch"):
        provider.decrypt(bad_envelope)

def test_decrypt_fail_tampered_ciphertext(provider):
    plaintext = b"data"
    envelope = provider.encrypt(plaintext)
    
    # Tamper ciphertext (flip last char)
    ct_bytes = bytearray(binascii.unhexlify(envelope.ciphertext))
    ct_bytes[-1] ^= 0xFF
    envelope.ciphertext = binascii.hexlify(ct_bytes).decode('ascii')
    
    # GCM should fail authentication
    from cryptography.exceptions import InvalidTag
    with pytest.raises(InvalidTag):
        provider.decrypt(envelope)

def test_envelope_serialization():
    env = Envelope(
        kek_id="k1",
        iv="00"*12,
        ciphertext="aa",
        tag="bb"*16,
        alg="aes-256-gcm"
    )
    json_str = env.to_json()
    d = json.loads(json_str)
    assert d["schema_id"] == SCHEMA_ID_ENVELOPE
    assert d["kek_id"] == "k1"
    
    env2 = Envelope.from_json(json_str)
    assert env2.kek_id == "k1"
    assert env2.iv == env.iv

def test_envelope_validation():
    from pydantic import ValidationError
    # Invalid IV
    with pytest.raises(ValidationError):
        Envelope(kek_id="k1", iv="too-short", ciphertext="aa", tag="bb"*16)
    
    # Invalid Tag
    with pytest.raises(ValidationError):
        Envelope(kek_id="k1", iv="00"*12, ciphertext="aa", tag="invalid-tag")



"""Cryptographic primitives for Talos SDK.

Provides wrappers for X25519, Ed25519, and common utilities.
"""

import base64
from typing import Any

from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from cryptography.hazmat.primitives import hashes
from pydantic import BaseModel, ConfigDict
import os


def b64u_encode(data: bytes) -> str:
    """Base64URL encoding without padding."""
    return base64.urlsafe_b64encode(data).decode("utf-8").rstrip("=")


def b64u_decode(data: str) -> bytes:
    """Base64URL decoding without padding."""
    padding = "=" * (4 - (len(data) % 4))
    if padding == "====":
        padding = ""
    return base64.urlsafe_b64decode(data + padding)


class KeyPair(BaseModel):
    """A generic keypair (Ed25519 or X25519)."""

    public_key: bytes
    private_key: bytes
    key_type: str = "x25519"  # x25519 or ed25519

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_dict(self) -> dict[str, Any]:
        """Convert to base64 dictionary."""
        return {
            "public": base64.b64encode(self.public_key).decode(),
            "private": base64.b64encode(self.private_key).decode(),
            "type": self.key_type,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "KeyPair":
        """Load from base64 dictionary."""
        return cls(
            public_key=base64.b64decode(data["public"]),
            private_key=base64.b64decode(data["private"]),
            key_type=data.get("type", "x25519"),
        )


def generate_encryption_keypair() -> KeyPair:
    """Generate an ephemeral X25519 keypair."""
    priv = X25519PrivateKey.generate()
    pub = priv.public_key()

    # Use Raw encoding for X25519
    from cryptography.hazmat.primitives.serialization import (
        Encoding,
        NoEncryption,
        PrivateFormat,
        PublicFormat,
    )

    pub_bytes = pub.public_bytes(encoding=Encoding.Raw, format=PublicFormat.Raw)
    priv_bytes = priv.private_bytes(
        encoding=Encoding.Raw,
        format=PrivateFormat.Raw,
        encryption_algorithm=NoEncryption(),
    )

    return KeyPair(
        public_key=pub_bytes,
        private_key=priv_bytes,
        key_type="x25519",
    )


def generate_signing_keypair() -> KeyPair:
    """Generate an Ed25519 signing keypair."""
    priv = Ed25519PrivateKey.generate()
    pub = priv.public_key()

    pub_bytes = pub.public_bytes_raw()
    priv_bytes = priv.private_bytes_raw()

    return KeyPair(
        public_key=pub_bytes,
        private_key=priv_bytes,
        key_type="ed25519",
    )


def sign_message(message: bytes, private_key_bytes: bytes) -> bytes:
    """Sign a message with Ed25519."""
    priv = Ed25519PrivateKey.from_private_bytes(private_key_bytes)
    return priv.sign(message)


def verify_signature(message: bytes, signature: bytes, public_key_bytes: bytes) -> bool:
    """Verify an Ed25519 signature."""
    try:
        pub = Ed25519PublicKey.from_public_bytes(public_key_bytes)
        pub.verify(signature, message)
        return True
    except Exception:
        return False


def derive_shared_secret(private_key_bytes: bytes, peer_public_key_bytes: bytes) -> bytes:
    """Derive X25519 shared secret."""
    priv = X25519PrivateKey.from_private_bytes(private_key_bytes)
    pub = X25519PublicKey.from_public_bytes(peer_public_key_bytes)
    return priv.exchange(pub)


def encrypt_message(message: bytes, key: bytes) -> tuple[bytes, bytes]:
    """Encrypt using ChaCha20Poly1305. Returns (nonce, ciphertext)."""
    cipher = ChaCha20Poly1305(key)
    nonce = os.urandom(12)
    ciphertext = cipher.encrypt(nonce, message, None)
    return nonce, ciphertext


def decrypt_message(ciphertext: bytes, key: bytes, nonce: bytes) -> bytes:
    """Decrypt using ChaCha20Poly1305."""
    cipher = ChaCha20Poly1305(key)
    return cipher.decrypt(nonce, ciphertext, None)


def hash_data(data: bytes) -> str:
    """Compute SHA-256 hash (hex string)."""
    digest = hashes.Hash(hashes.SHA256())
    digest.update(data)
    return digest.finalize().hex()



import xml.etree.ElementTree as ET


class JUnitReport:
    def __init__(self) -> None:
        self.testsuites: list[ET.Element] = []

    def add_testsuite(
        self,
        name: str,
        tests: int = 0,
        failures: int = 0,
        errors: int = 0,
        time: float = 0.0,
    ) -> ET.Element:
        suite = ET.Element(
            "testsuite",
            {
                "name": name,
                "tests": str(tests),
                "failures": str(failures),
                "errors": str(errors),
                "time": f"{time:.4f}",
            },
        )
        self.testsuites.append(suite)
        return suite

    def add_testcase(
        self, suite: ET.Element, name: str, classname: str, time: float = 0.0
    ) -> ET.Element:
        case = ET.SubElement(
            suite,
            "testcase",
            {"name": name, "classname": classname, "time": f"{time:.4f}"},
        )
        return case

    def add_failure(
        self, case: ET.Element, message: str, type: str = "AssertionError"
    ) -> None:
        failure = ET.SubElement(case, "failure", {"message": message, "type": type})
        failure.text = message

    def add_error(self, case: ET.Element, message: str, type: str = "Error") -> None:
        error = ET.SubElement(case, "error", {"message": message, "type": type})
        error.text = message

    def write(self, path: str) -> None:
        root = ET.Element("testsuites")
        for suite in self.testsuites:
            root.append(suite)

        tree = ET.ElementTree(root)
        ET.indent(tree, space="  ", level=0)
        tree.write(path, encoding="utf-8", xml_declaration=True)



"""Talos SDK - Secure Multi-Language SDK for AI Agents.

This package provides the Python implementation of the Talos Protocol v1,
enabling secure identity, capability authorization, and MCP request signing.

Example:
    >>> from talos_sdk import Wallet, TalosClient
    >>> wallet = Wallet.generate("my-agent")
    >>> client = TalosClient("wss://gateway.example.com", wallet)
    >>> await client.connect()
    >>> response = await client.sign_and_send_mcp(
    ...     {"method": "read", "params": {"path": "/data"}},
    ...     "filesystem",
    ...     "read",
    ... )
"""

from .canonical import canonical_json, canonical_json_bytes
from .client import PROTOCOL_VERSION, TalosClient
from .errors import (
    TalosCryptoError,
    TalosDeniedError,
    TalosError,
    TalosFrameInvalidError,
    TalosInvalidCapabilityError,
    TalosInvalidInputError,
    TalosProtocolMismatchError,
    TalosTransportError,
    TalosTransportTimeoutError,
)
from .mcp import SignedFrame, sign_mcp_request, verify_mcp_response
from .mcp_client import McpClient
from .wallet import Wallet

__version__ = "1.0.0"

# SDK Version Exports (required by VERSIONING.md)
SDK_VERSION = __version__
SUPPORTED_PROTOCOL_RANGE = ("1.0", "1.x")
CONTRACT_MANIFEST_HASH = "sha256:pending"  # Updated at build time

__all__ = [
    # Core classes
    "Wallet",
    "TalosClient",
    "McpClient",
    "SignedFrame",
    # Functions
    "sign_mcp_request",
    "verify_mcp_response",
    "canonical_json",
    "canonical_json_bytes",
    # Constants
    "PROTOCOL_VERSION",
    # Errors
    "TalosError",
    "TalosDeniedError",
    "TalosInvalidCapabilityError",
    "TalosProtocolMismatchError",
    "TalosFrameInvalidError",
    "TalosCryptoError",
    "TalosInvalidInputError",
    "TalosTransportTimeoutError",
    "TalosTransportError",
]



from talos_sdk.crypto import (
    KeyPair,
    b64u_decode,
    b64u_encode,
    generate_encryption_keypair,
    generate_signing_keypair,
    sign_message,
    verify_signature,
)


def test_b64u_encoding():
    data = b"hello world"
    encoded = b64u_encode(data)
    assert encoded == "aGVsbG8gd29ybGQ"  # no padding
    decoded = b64u_decode(encoded)
    assert decoded == data


def test_b64u_decode_padding():
    # Test cases with different lengths requiring padding
    assert b64u_decode("YQ") == b"a"  # "YQ=="
    assert b64u_decode("YWI") == b"ab"  # "YWI="
    assert b64u_decode("YWJj") == b"abc"  # "YWJj"


def test_encryption_keypair_lifecycle():
    kp = generate_encryption_keypair()
    assert kp.key_type == "x25519"
    assert len(kp.public_key) == 32
    assert len(kp.private_key) == 32

    # Test serialization
    data = kp.to_dict()
    assert data["type"] == "x25519"
    assert "public" in data
    assert "private" in data

    # Test deserialization
    kp2 = KeyPair.from_dict(data)
    assert kp2.public_key == kp.public_key
    assert kp2.private_key == kp.private_key
    assert kp2.key_type == kp.key_type


def test_signing_keypair_lifecycle():
    kp = generate_signing_keypair()
    assert kp.key_type == "ed25519"
    assert len(kp.public_key) == 32
    assert len(kp.private_key) == 32


def test_sign_and_verify():
    kp = generate_signing_keypair()
    msg = b"test message for signing"

    # Sign
    signature = sign_message(msg, kp.private_key)
    assert len(signature) == 64

    # Verify Success
    assert verify_signature(msg, signature, kp.public_key) is True

    # Verify Failure (Wrong Message)
    assert verify_signature(b"altered message", signature, kp.public_key) is False

    # Verify Failure (Wrong Key)
    other_kp = generate_signing_keypair()
    assert verify_signature(msg, signature, other_kp.public_key) is False


def test_verify_invalid_input():
    # Helper to test exception handling in verify_signature
    assert verify_signature(b"msg", b"sig", b"short_key") is False



"""Double Ratchet Protocol for Perfect Forward Secrecy.

This module implements the Signal Double Ratchet algorithm for secure
messaging with forward secrecy and break-in recovery.
"""

import base64
import json
import logging
import os
import time
from typing import Any

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from pydantic import BaseModel, ConfigDict, Field, field_serializer

from .canonical import canonical_json_bytes
from .crypto import (
    KeyPair,
    b64u_decode,
    b64u_encode,
    generate_encryption_keypair,
    sign_message,
    verify_signature,
)
from .errors import TalosError

logger = logging.getLogger(__name__)

# Protocol constants
MAX_SKIP = 1000  # Max messages to skip in a chain
INFO_ROOT = b"talos-double-ratchet-root"
INFO_CHAIN = b"talos-double-ratchet-chain"
INFO_MESSAGE = b"talos-double-ratchet-message"


class RatchetError(TalosError):
    """Error during ratchet operation."""

    def __init__(self, message: str, **kwargs: Any) -> None:
        super().__init__("TALOS_RATCHET_ERROR", message, **kwargs)


class PrekeyBundle(BaseModel):
    """Prekey bundle for X3DH key exchange.

    Published by users to allow others to establish sessions.
    """

    identity_key: bytes  # Long-term Ed25519 public key
    signed_prekey: bytes  # X25519 public key, signed by identity key
    prekey_signature: bytes  # Signature over signed_prekey
    one_time_prekey: bytes | None = None  # Optional ephemeral X25519 key

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("identity_key", "signed_prekey", "prekey_signature")
    def serialize_bytes(self, v: bytes, _info: Any) -> str:
        return base64.b64encode(v).decode()

    @field_serializer("one_time_prekey")
    def serialize_opt_bytes(self, v: bytes | None, _info: Any) -> str | None:
        if v is None:
            return None
        return base64.b64encode(v).decode()

    def verify(self) -> bool:
        """Verify the prekey signature."""
        return verify_signature(
            self.signed_prekey, self.prekey_signature, self.identity_key
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary with base64-encoded keys (compat alias)."""
        return self.model_dump()

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "PrekeyBundle":
        return cls(
            identity_key=base64.b64decode(data["identity_key"]),
            signed_prekey=base64.b64decode(data["signed_prekey"]),
            prekey_signature=base64.b64decode(data["prekey_signature"]),
            one_time_prekey=base64.b64decode(data["one_time_prekey"])
            if data.get("one_time_prekey")
            else None,
        )


class MessageHeader(BaseModel):
    """Header for ratcheted messages.

    Contains the sender's current DH public key and chain position.
    """

    dh_public: bytes  # Sender's current DH ratchet public key
    previous_chain_length: int  # Messages in previous sending chain
    message_number: int  # Index in current sending chain

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("dh_public")
    def serialize_bytes(self, v: bytes, _info: Any) -> str:
        return base64.b64encode(v).decode()

    def to_bytes(self) -> bytes:
        return canonical_json_bytes(
            {
                "dh": b64u_encode(self.dh_public),
                "pn": self.previous_chain_length,
                "n": self.message_number,
            }
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> "MessageHeader":
        d = json.loads(data)
        return cls(
            dh_public=b64u_decode(d["dh"]),
            previous_chain_length=d["pn"],
            message_number=d["n"],
        )


class RatchetState(BaseModel):
    """State of a Double Ratchet session.

    This contains all the keys and counters needed to encrypt
    and decrypt messages with forward secrecy.
    """

    # DH ratchet keys
    dh_keypair: KeyPair  # Our current DH key pair
    dh_remote: bytes | None  # Remote's current DH public key

    # Root key (updated on DH ratchet)
    root_key: bytes

    # Sending and receiving chain keys
    chain_key_send: bytes | None = None
    chain_key_recv: bytes | None = None

    # Message counters
    send_count: int = 0  # Messages sent in current sending chain
    recv_count: int = 0  # Messages received in current receiving chain
    prev_send_count: int = 0  # Messages in previous sending chain

    # Skipped message keys (for out-of-order delivery)
    skipped_keys: dict[tuple[bytes, int], bytes] = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("dh_remote", "root_key", "chain_key_send", "chain_key_recv")
    def serialize_opt_bytes(self, v: bytes | None, _info: Any) -> str | None:
        if v is None:
            return None
        return base64.b64encode(v).decode()

    @field_serializer("dh_keypair")
    def serialize_keypair(self, v: KeyPair, _info: Any) -> dict[str, Any]:
        return v.to_dict()

    @field_serializer("skipped_keys")
    def serialize_skipped(
        self, v: dict[tuple[bytes, int], bytes], _info: Any
    ) -> list[dict[str, Any]]:
        return [
            {
                "dh": base64.b64encode(k[0]).decode(),
                "n": k[1],
                "key": base64.b64encode(val).decode(),
            }
            for k, val in v.items()
        ]

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary (compat alias)."""
        return self.model_dump()

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "RatchetState":
        skipped = {}
        for item in data.get(
            "skipped_keys", data.get("skipped", [])
        ):  # Handle both key names for compat
            key = (base64.b64decode(item["dh"]), item["n"])
            skipped[key] = base64.b64decode(item["key"])

        return cls(
            dh_keypair=KeyPair.from_dict(data["dh_keypair"]),
            dh_remote=base64.b64decode(data["dh_remote"])
            if data.get("dh_remote")
            else None,
            root_key=base64.b64decode(data["root_key"]),
            chain_key_send=base64.b64decode(data["chain_key_send"])
            if data.get("chain_key_send")
            else None,
            chain_key_recv=base64.b64decode(data["chain_key_recv"])
            if data.get("chain_key_recv")
            else None,
            send_count=data.get("send_count", 0),
            recv_count=data.get("recv_count", 0),
            prev_send_count=data.get("prev_send_count", 0),
            skipped_keys=skipped,
        )


def _hkdf_derive(input_key: bytes, info: bytes, length: int = 32) -> bytes:
    """Derive key using HKDF-SHA256."""
    hkdf = HKDF(
        algorithm=hashes.SHA256(),
        length=length,
        salt=None,
        info=info,
    )
    return hkdf.derive(input_key)


def _kdf_rk(rk: bytes, dh_out: bytes) -> tuple[bytes, bytes]:
    """Root key KDF: derive new root key and chain key.

    Returns (new_root_key, new_chain_key)
    """
    # Derive 64 bytes from root key + DH output
    combined = _hkdf_derive(rk + dh_out, INFO_ROOT, length=64)
    return combined[:32], combined[32:]


def _kdf_ck(ck: bytes) -> tuple[bytes, bytes]:
    """Chain key KDF: derive message key and next chain key.

    Returns (message_key, next_chain_key)
    """
    # Use HMAC-based approach (simplified HKDF)
    message_key = _hkdf_derive(ck, INFO_MESSAGE)
    next_chain_key = _hkdf_derive(ck, INFO_CHAIN)
    return message_key, next_chain_key


def _dh(private: bytes, public: bytes) -> bytes:
    """Perform X25519 Diffie-Hellman exchange."""
    priv = X25519PrivateKey.from_private_bytes(private)
    pub = X25519PublicKey.from_public_bytes(public)
    return priv.exchange(pub)


def _encrypt_aead(key: bytes, plaintext: bytes, ad: bytes) -> tuple[bytes, bytes]:
    """Encrypt with ChaCha20-Poly1305 AEAD. Returns (nonce, ciphertext)."""
    nonce = os.urandom(12)
    cipher = ChaCha20Poly1305(key)
    ciphertext = cipher.encrypt(nonce, plaintext, ad)
    return nonce, ciphertext


def _decrypt_aead(key: bytes, nonce: bytes, ciphertext: bytes, ad: bytes) -> bytes:
    """Decrypt with ChaCha20-Poly1305 AEAD."""
    cipher = ChaCha20Poly1305(key)
    return cipher.decrypt(nonce, ciphertext, ad)


class Session:
    """A Double Ratchet session with a single peer.

    Provides forward-secure encryption with per-message keys.
    """

    def __init__(self, peer_id: str, state: RatchetState) -> None:
        self.peer_id = peer_id
        self.state = state
        self.created_at = time.time()
        self.last_activity = time.time()
        self.messages_sent = 0
        self.messages_received = 0

    def encrypt(self, plaintext: bytes) -> bytes:
        """Encrypt a message with the current sending key.

        Returns JSON-enveloped message as canonical bytes.
        """
        if self.state.chain_key_send is None:
            self._initialize_sending_chain()

        assert self.state.chain_key_send is not None
        mk, self.state.chain_key_send = _kdf_ck(self.state.chain_key_send)

        header = MessageHeader(
            dh_public=self.state.dh_keypair.public_key,
            previous_chain_length=self.state.prev_send_count,
            message_number=self.state.send_count,
        )

        # Encrypt with raw header bytes as associated data
        header_bytes = header.to_bytes()
        nonce, ciphertext = _encrypt_aead(mk, plaintext, header_bytes)

        # Update counters
        self.state.send_count += 1
        self.messages_sent += 1
        self.last_activity = time.time()

        # Wrap in JSON envelope (v1.1.0 Option A)
        envelope = {
            "header": {
                "dh": b64u_encode(header.dh_public),
                "pn": header.previous_chain_length,
                "n": header.message_number,
            },
            "nonce": b64u_encode(nonce),
            "ciphertext": b64u_encode(ciphertext),
        }
        return canonical_json_bytes(envelope)

    def decrypt(self, message: bytes) -> bytes:
        """Decrypt a message, performing DH ratchet if needed."""
        # Parse envelope
        try:
            envelope = json.loads(message)
            header_dict = envelope["header"]
            # Re-canonicalize header for AAD and parsing
            header_bytes = canonical_json_bytes(header_dict)
            header = MessageHeader.from_bytes(header_bytes)
            nonce = b64u_decode(envelope["nonce"])
            ciphertext = b64u_decode(envelope["ciphertext"])
        except (KeyError, ValueError) as e:
            raise RatchetError(f"Malformed message envelope: {e}")

        # Try skipped keys first
        plaintext = self._try_skipped_keys(header, nonce, ciphertext, header_bytes)
        if plaintext is not None:
            return plaintext

        # Check if we need a DH ratchet
        if header.dh_public != self.state.dh_remote:
            # Skip messages in the old receiving chain
            self._skip_message_keys(header.previous_chain_length)
            # Perform DH ratchet
            self._dh_ratchet(header)

        # Skip any messages in current receiving chain
        self._skip_message_keys(header.message_number)

        # Decrypt with current key
        if self.state.chain_key_recv is None:
            raise RatchetError("No receiving chain key")

        mk, self.state.chain_key_recv = _kdf_ck(self.state.chain_key_recv)
        self.state.recv_count += 1

        try:
            plaintext = _decrypt_aead(mk, nonce, ciphertext, header_bytes)
        except Exception as e:
            raise RatchetError(f"Decryption failed: {e}")

        self.messages_received += 1
        self.last_activity = time.time()

        return plaintext

    def _try_skipped_keys(
        self,
        header: MessageHeader,
        nonce: bytes,
        ciphertext: bytes,
        ad: bytes,
    ) -> bytes | None:
        """Try to decrypt with a skipped message key."""
        key_id = (header.dh_public, header.message_number)
        if key_id in self.state.skipped_keys:
            mk = self.state.skipped_keys.pop(key_id)
            return _decrypt_aead(mk, nonce, ciphertext, ad)
        return None

    def _skip_message_keys(self, until: int) -> None:
        """Store skipped message keys for out-of-order messages."""
        if self.state.chain_key_recv is None:
            return

        if self.state.recv_count + MAX_SKIP < until:
            raise RatchetError("Too many skipped messages")

        while self.state.recv_count < until:
            mk, self.state.chain_key_recv = _kdf_ck(self.state.chain_key_recv)
            assert self.state.dh_remote is not None
            key_id = (self.state.dh_remote, self.state.recv_count)
            self.state.skipped_keys[key_id] = mk
            self.state.recv_count += 1

    def _dh_ratchet(self, header: MessageHeader) -> None:
        """Perform a DH ratchet step."""
        self.state.prev_send_count = self.state.send_count
        self.state.send_count = 0
        self.state.recv_count = 0
        self.state.dh_remote = header.dh_public

        # Derive new receiving chain
        assert self.state.dh_remote is not None
        dh_recv = _dh(self.state.dh_keypair.private_key, self.state.dh_remote)
        self.state.root_key, self.state.chain_key_recv = _kdf_rk(
            self.state.root_key, dh_recv
        )

        # Generate new DH key pair
        self.state.dh_keypair = generate_encryption_keypair()

        # Derive new sending chain
        dh_send = _dh(self.state.dh_keypair.private_key, self.state.dh_remote)
        self.state.root_key, self.state.chain_key_send = _kdf_rk(
            self.state.root_key, dh_send
        )

    def _initialize_sending_chain(self) -> None:
        """Initialize a new sending chain."""
        self.state.prev_send_count = self.state.send_count
        self.state.send_count = 0

        # Generate new DH key pair
        self.state.dh_keypair = generate_encryption_keypair()

        # Derive new sending chain
        # Derive new sending chain
        assert self.state.dh_remote is not None
        dh_send = _dh(self.state.dh_keypair.private_key, self.state.dh_remote)
        self.state.root_key, self.state.chain_key_send = _kdf_rk(
            self.state.root_key, dh_send
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "peer_id": self.peer_id,
            "state": self.state.to_dict(),
            "created_at": self.created_at,
            "last_activity": self.last_activity,
            "messages_sent": self.messages_sent,
            "messages_received": self.messages_received,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Session":
        session = cls(
            peer_id=data["peer_id"],
            state=RatchetState.from_dict(data["state"]),
        )
        session.created_at = data.get("created_at", time.time())
        session.last_activity = data.get("last_activity", time.time())
        session.messages_sent = data.get("messages_sent", 0)
        session.messages_received = data.get("messages_received", 0)
        return session


class SessionManager:
    """Manages Double Ratchet sessions with multiple peers."""

    def __init__(
        self, identity_keypair: KeyPair, storage_path: str | None = None
    ) -> None:
        self.identity_keypair = identity_keypair
        self.storage_path = storage_path
        self.sessions: dict[str, Session] = {}

        # Our prekey for others to contact us
        self._signed_prekey = generate_encryption_keypair()
        self._prekey_signature = sign_message(
            self._signed_prekey.public_key, identity_keypair.private_key
        )

    def get_prekey_bundle(self) -> PrekeyBundle:
        """Get our prekey bundle for publishing."""
        return PrekeyBundle(
            identity_key=self.identity_keypair.public_key,
            signed_prekey=self._signed_prekey.public_key,
            prekey_signature=self._prekey_signature,
        )

    def create_session_as_initiator(
        self,
        peer_id: str,
        peer_bundle: PrekeyBundle,
    ) -> Session:
        """Create a new session as the initiator (Alice)."""
        # Verify peer's prekey signature
        if not peer_bundle.verify():
            raise RatchetError("Invalid prekey signature")

        # Generate ephemeral key (will be reused as first ratchet key)
        dh_keypair = generate_encryption_keypair()

        # X3DH: Compute shared secret
        dh_x3dh = _dh(dh_keypair.private_key, peer_bundle.signed_prekey)

        # Derive initial root key
        root_key = _hkdf_derive(dh_x3dh, b"x3dh-init")

        # Initialize first sending chain
        dh_out = _dh(dh_keypair.private_key, peer_bundle.signed_prekey)
        root_key, chain_key_send = _kdf_rk(root_key, dh_out)

        state = RatchetState(
            dh_keypair=dh_keypair,
            dh_remote=peer_bundle.signed_prekey,
            root_key=root_key,
            chain_key_send=chain_key_send,
        )

        session = Session(peer_id, state)
        self.sessions[peer_id] = session
        return session

    def create_session_as_responder(
        self,
        peer_id: str,
        peer_dh_public: bytes,
    ) -> Session:
        """Create a new session as the responder (Bob)."""
        dh_x3dh = _dh(self._signed_prekey.private_key, peer_dh_public)
        root_key = _hkdf_derive(dh_x3dh, b"x3dh-init")

        dh_recv = _dh(self._signed_prekey.private_key, peer_dh_public)
        root_key, chain_key_recv = _kdf_rk(root_key, dh_recv)

        state = RatchetState(
            dh_keypair=self._signed_prekey,
            dh_remote=peer_dh_public,
            root_key=root_key,
            chain_key_recv=chain_key_recv,
        )

        session = Session(peer_id, state)
        self.sessions[peer_id] = session
        return session

    def get_session(self, peer_id: str) -> Session | None:
        return self.sessions.get(peer_id)



"""A2A Session Client facade with FrameCrypto hook for Phase 10.3."""

import hashlib
from datetime import datetime, timezone
from typing import Protocol

from talos_sdk.canonical import canonical_json_bytes

from .errors import A2ACryptoNotConfiguredError
from .models import EncryptedFrame, FrameListResponse, FrameSendResponse
from .sequence_tracker import SequenceStorage, SequenceTracker
from .transport import A2ATransport


class FrameCrypto(Protocol):
    """Hook for Phase 10.3 Ratchet Binding.

    Implement this protocol to enable send_message() and receive_messages().
    """

    def encrypt(self, plaintext: bytes) -> tuple[str, str, str]:
        """Encrypt plaintext and return frame components.

        Returns:
            Tuple of (header_b64u, ciphertext_b64u, ciphertext_hash)

        Note:
            frame_digest is computed by session_client after reserving sender_seq.
        """
        ...

    def decrypt(
        self,
        header_b64u: str,
        ciphertext_b64u: str,
        ciphertext_hash: str | None = None,
    ) -> bytes:
        """Decrypt frame and return plaintext."""
        ...


class A2ASessionClient:
    """High-level session facade for A2A communication.

    Combines transport, sequence tracking, and optional crypto.

    Phase 10.2: Transport-only. Use send_frame() and receive_frames().
    Phase 10.3: With crypto hook, use send_message() and receive_messages().

    Usage:
        # As initiator
        client = await A2ASessionClient.initiate(transport, sender_id, responder_id)

        # As responder
        client = await A2ASessionClient.accept(transport, session_id, sender_id, initiator_id)

        # Send frame (Phase 10.2)
        await client.send_frame(encrypted_frame)

        # Receive frames
        resp = await client.receive_frames()
    """

    def __init__(
        self,
        transport: A2ATransport,
        session_id: str,
        sender_id: str,
        peer_id: str,
        is_initiator: bool,
        sequence_storage: SequenceStorage | None = None,
        crypto: FrameCrypto | None = None,
    ):
        self._transport = transport
        self._session_id = session_id
        self._sender_id = sender_id
        self._peer_id = peer_id
        self._is_initiator = is_initiator
        self._seq_tracker = SequenceTracker(session_id, sender_id, sequence_storage)
        self._crypto = crypto

    @property
    def session_id(self) -> str:
        """Get the session ID."""
        return self._session_id

    @property
    def sender_id(self) -> str:
        """Get the sender (self) ID."""
        return self._sender_id

    @property
    def peer_id(self) -> str:
        """Get the peer ID."""
        return self._peer_id

    @property
    def is_initiator(self) -> bool:
        """Whether this client initiated the session."""
        return self._is_initiator

    async def send_frame(self, frame: EncryptedFrame) -> FrameSendResponse:
        """Send a pre-constructed encrypted frame.

        Use this in Phase 10.2 when constructing frames externally.
        """
        return await self._transport.send_frame(self._session_id, frame)

    async def receive_frames(
        self, cursor: str | None = None, limit: int = 100
    ) -> FrameListResponse:
        """Receive encrypted frames with cursor pagination."""
        return await self._transport.receive_frames(
            self._session_id, cursor=cursor, limit=limit
        )

    async def send_message(self, plaintext: bytes) -> FrameSendResponse:
        """Encrypt and send a message.

        Requires FrameCrypto hook (Phase 10.3).
        Sequence is reserved atomically before encryption for retry safety.

        Raises:
            A2ACryptoNotConfiguredError: If crypto hook not provided.
        """
        if self._crypto is None:
            raise A2ACryptoNotConfiguredError()

        # Reserve sequence ONCE before encryption (retry-safe)
        sender_seq = self._seq_tracker.reserve()

        # Encrypt using crypto hook (returns 3 values, not 4)
        header_b64u, ciphertext_b64u, ciphertext_hash = self._crypto.encrypt(plaintext)

        # Compute frame_digest per LOCKED SPEC
        # Preimage: {schema_id, schema_version, session_id, sender_id, sender_seq, header_b64u, ciphertext_hash}
        preimage = {
            "schema_id": "talos.a2a.encrypted_frame",
            "schema_version": "v1",
            "session_id": self._session_id,
            "sender_id": self._sender_id,
            "sender_seq": sender_seq,
            "header_b64u": header_b64u,
            "ciphertext_hash": ciphertext_hash,
        }
        frame_digest = hashlib.sha256(canonical_json_bytes(preimage)).hexdigest()

        frame = EncryptedFrame(
            session_id=self._session_id,
            sender_id=self._sender_id,
            sender_seq=sender_seq,
            header_b64u=header_b64u,
            ciphertext_b64u=ciphertext_b64u,
            frame_digest=frame_digest,
            ciphertext_hash=ciphertext_hash,
            created_at=datetime.now(timezone.utc),
        )

        # Transport handles retry internally (same frame on timeout)
        return await self.send_frame(frame)

    async def receive_messages(self, cursor: str | None = None) -> list[bytes]:
        """Receive and decrypt messages.

        Requires FrameCrypto hook (Phase 10.3).

        Raises:
            A2ACryptoNotConfiguredError: If crypto hook not provided.
        """
        if self._crypto is None:
            raise A2ACryptoNotConfiguredError()

        resp = await self.receive_frames(cursor)
        return [
            self._crypto.decrypt(
                frame.header_b64u, frame.ciphertext_b64u, frame.ciphertext_hash
            )
            for frame in resp.items
        ]

    async def rotate(self) -> None:
        """Rotate session keys."""
        await self._transport.rotate_session(self._session_id)

    async def close(self) -> None:
        """Close the session."""
        await self._transport.close_session(self._session_id)

    @classmethod
    async def initiate(
        cls,
        transport: A2ATransport,
        sender_id: str,
        responder_id: str,
        sequence_storage: SequenceStorage | None = None,
        crypto: FrameCrypto | None = None,
    ) -> "A2ASessionClient":
        """Create a new session as initiator."""
        resp = await transport.create_session(responder_id)
        return cls(
            transport=transport,
            session_id=resp.session_id,
            sender_id=sender_id,
            peer_id=responder_id,
            is_initiator=True,
            sequence_storage=sequence_storage,
            crypto=crypto,
        )

    @classmethod
    async def accept(
        cls,
        transport: A2ATransport,
        session_id: str,
        sender_id: str,
        initiator_id: str,
        sequence_storage: SequenceStorage | None = None,
        crypto: FrameCrypto | None = None,
    ) -> "A2ASessionClient":
        """Accept a pending session as responder."""
        await transport.accept_session(session_id)
        return cls(
            transport=transport,
            session_id=session_id,
            sender_id=sender_id,
            peer_id=initiator_id,
            is_initiator=False,
            sequence_storage=sequence_storage,
            crypto=crypto,
        )



from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)

from talos_sdk.ports.crypto import ICryptoPort


class Ed25519CryptoAdapter(ICryptoPort):
    def sign(self, data: bytes, key: bytes) -> bytes:
        """Sign data using Ed25519 private key.
        Key must be 32 bytes (seed).
        """
        # Load from seed (32 bytes)
        private_key = Ed25519PrivateKey.from_private_bytes(key)
        return private_key.sign(data)

    def verify(self, data: bytes, sig: bytes, key: bytes) -> bool:
        """Verify signature using Ed25519 public key.
        Key must be 32 bytes.
        """
        try:
            public_key = Ed25519PublicKey.from_public_bytes(key)
            public_key.verify(sig, data)
            return True
        except InvalidSignature:
            return False
        except ValueError:
            # Invalid key format
            return False



from abc import ABC, abstractmethod
from typing import Any


class IHashPort(ABC):
    """Port for hashing operations.
    Includes standard hashing and canonical object hashing.
    """

    @abstractmethod
    def sha256(self, data: bytes) -> bytes:
        """Compute SHA-256 hash of bytes."""
        ...

    @abstractmethod
    def canonical_hash(self, obj: Any) -> bytes:
        """Compute hash of an object using canonical JSON Serialization.
        Essential for content-addressable storage and signatures.
        """
        ...



import os
import sys
from dataclasses import dataclass

# Ensure src is in path
sys.path.append(os.path.join(os.path.dirname(__file__), "../src"))

from talos_sdk.adapters.crypto import Ed25519CryptoAdapter  # noqa: E402
from talos_sdk.adapters.hash import NativeHashAdapter  # noqa: E402
from talos_sdk.adapters.memory_store import (  # noqa: E402
    InMemoryAuditStore,
    InMemoryKeyValueStore,
)


def test_hash_adapter():
    adapter = NativeHashAdapter()
    data = b"hello"
    h = adapter.sha256(data)
    assert len(h) == 32
    # Known hash for "hello"
    assert h.hex() == "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"

    obj = {"b": 2, "a": 1}
    # Canonical: {"a":1,"b":2}
    h_canon = adapter.canonical_hash(obj)
    # sha256({"a":1,"b":2})
    # echo -n '{"a":1,"b":2}' | openssl dgst -sha256
    # 6325...
    assert len(h_canon) == 32


def test_crypto_adapter():
    adapter = Ed25519CryptoAdapter()
    # 32 bytes seed
    seed = b"12345678901234567890123456789012"
    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

    priv = Ed25519PrivateKey.from_private_bytes(seed)
    # pub bytes
    from cryptography.hazmat.primitives import serialization

    pub_bytes = priv.public_key().public_bytes(
        encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw
    )
    # Actually just call sign
    msg = b"hello world"
    sig = adapter.sign(msg, seed)
    assert len(sig) == 64

    valid = adapter.verify(msg, sig, pub_bytes)
    assert valid is True

    invalid = adapter.verify(b"hallo", sig, pub_bytes)
    assert invalid is False


def test_kv_store():
    store = InMemoryKeyValueStore()
    store.put("k", b"v")
    assert store.get("k") == b"v"
    store.delete("k")
    assert store.get("k") is None


@dataclass
class MockEvent:
    event_id: str
    timestamp: float


def test_audit_store():
    store = InMemoryAuditStore()
    e1 = MockEvent("1", 100.0)
    store.append(e1)

    page = store.list()
    assert len(page.events) == 1
    assert page.events[0] == e1



"""Talos SDK A2A Module.

Provides HTTP transport, sequence tracking, and session management
for secure Agent-to-Agent communication.

Phase 10.2: Transport-only (no encryption). Phase 10.3 adds FrameCrypto binding.
"""

from .errors import (
    A2ACryptoNotConfiguredError,
    A2AError,
    A2AFrameDigestMismatchError,
    A2AFrameReplayError,
    A2AFrameSequenceTooFarError,
    A2ASessionNotFoundError,
    A2ASessionStateInvalidError,
    A2ATransportError,
)
from .models import (
    EncryptedFrame,
    ErrorResponse,
    FrameListResponse,
    FrameSendResponse,
    GroupResponse,
    SessionResponse,
)
from .ratchet_crypto import RatchetFrameCrypto
from .sequence_tracker import InMemorySequenceStorage, SequenceStorage, SequenceTracker
from .session_client import A2ASessionClient, FrameCrypto
from .transport import A2ATransport

__all__ = [
    # Errors
    "A2AError",
    "A2ATransportError",
    "A2ASessionNotFoundError",
    "A2ASessionStateInvalidError",
    "A2AFrameReplayError",
    "A2AFrameDigestMismatchError",
    "A2AFrameSequenceTooFarError",
    "A2ACryptoNotConfiguredError",
    # Models
    "EncryptedFrame",
    "SessionResponse",
    "FrameSendResponse",
    "FrameListResponse",
    "GroupResponse",
    "ErrorResponse",
    # Sequence
    "SequenceTracker",
    "SequenceStorage",
    "InMemorySequenceStorage",
    # Transport
    "A2ATransport",
    # Session
    "A2ASessionClient",
    "FrameCrypto",
    # Ratchet
    "RatchetFrameCrypto",
]



"""
Tests for the Talos SDK Wallet module.
"""

import pytest

from talos_sdk import TalosInvalidInputError, Wallet


class TestWallet:
    """Tests for Wallet class."""

    def test_generate_creates_valid_wallet(self):
        """Generate should create a valid wallet."""
        wallet = Wallet.generate(name="test")
        assert wallet is not None
        assert wallet.name == "test"
        assert len(wallet.public_key) == 32
        assert len(wallet.address) == 64  # SHA256 hex

    def test_generate_produces_unique_wallets(self):
        """Each generate call should produce a unique wallet."""
        wallet1 = Wallet.generate()
        wallet2 = Wallet.generate()
        assert wallet1.public_key != wallet2.public_key

    def test_from_seed_is_deterministic(self):
        """from_seed should be deterministic."""
        seed = bytes(32)  # All zeros
        wallet1 = Wallet.from_seed(seed)
        wallet2 = Wallet.from_seed(seed)
        assert wallet1.public_key == wallet2.public_key
        assert wallet1.to_did() == wallet2.to_did()

    def test_from_seed_rejects_wrong_length(self):
        """from_seed should reject seeds that aren't 32 bytes."""
        with pytest.raises(TalosInvalidInputError) as exc:
            Wallet.from_seed(bytes(16))
        assert "TALOS_INVALID_INPUT" == exc.value.code
        assert "32 bytes" in exc.value.message

    def test_to_did_format(self):
        """to_did should return a valid DID string."""
        wallet = Wallet.generate()
        did = wallet.to_did()
        assert did.startswith("did:key:z")
        assert len(did) > 20

    def test_sign_produces_64_byte_signature(self):
        """sign should produce a 64-byte Ed25519 signature."""
        wallet = Wallet.generate()
        signature = wallet.sign(b"test message")
        assert len(signature) == 64

    def test_sign_is_deterministic(self):
        """Ed25519 signing should be deterministic."""
        seed = bytes(32)
        wallet = Wallet.from_seed(seed)
        sig1 = wallet.sign(b"test")
        sig2 = wallet.sign(b"test")
        assert sig1 == sig2

    def test_verify_valid_signature(self):
        """verify should return True for valid signature."""
        wallet = Wallet.generate()
        message = b"test message"
        signature = wallet.sign(message)
        assert Wallet.verify(message, signature, wallet.public_key) is True

    def test_verify_invalid_signature(self):
        """verify should return False for invalid signature."""
        wallet = Wallet.generate()
        message = b"test message"
        # Use a bad signature (all zeros)
        bad_sig = bytes(64)
        assert Wallet.verify(message, bad_sig, wallet.public_key) is False

    def test_verify_wrong_message(self):
        """verify should return False for wrong message."""
        wallet = Wallet.generate()
        signature = wallet.sign(b"original")
        assert Wallet.verify(b"tampered", signature, wallet.public_key) is False

    def test_verify_wrong_key(self):
        """verify should return False for wrong public key."""
        wallet1 = Wallet.generate()
        wallet2 = Wallet.generate()
        message = b"test"
        signature = wallet1.sign(message)
        assert Wallet.verify(message, signature, wallet2.public_key) is False



from collections.abc import Callable
from typing import Any, TypeVar, cast

T = TypeVar("T")


class Container:
    """A simple thread-safe Dependency Injection container."""

    def __init__(self) -> None:
        self._services: dict[type[Any], Any] = {}
        self._factories: dict[type[Any], Callable[[], Any]] = {}

    def register(self, interface: type[T], implementation: T) -> None:
        """Register a singleton instance for an interface."""
        self._services[interface] = implementation

    def register_factory(self, interface: type[T], factory: Callable[[], T]) -> None:
        """Register a factory for an interface (lazy or transient)."""
        self._factories[interface] = factory

    def resolve(self, interface: type[T]) -> T:
        """Resolve an implementation for an interface."""
        if interface in self._services:
            return cast(T, self._services[interface])

        if interface in self._factories:
            instance = self._factories[interface]()
            # Optionally cache if singleton desired, but keeping simple
            return cast(T, instance)

        raise ValueError(f"No registration found for {interface}")


# Global instance for easy access if needed
_global_container = Container()


def get_container() -> Container:
    return _global_container



"""
Tests for the Talos SDK TalosClient.
"""

import pytest

from talos_sdk import TalosClient, TalosTransportError, Wallet


class TestTalosClient:
    """Tests for TalosClient facade."""

    def test_client_creation(self):
        """TalosClient should be creatable with wallet."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)
        assert client.wallet is wallet

    def test_protocol_version(self):
        """protocol_version should return version string."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)
        version = client.protocol_version()
        assert version == "1.0"

    def test_supported_protocol_range(self):
        """supported_protocol_range should return min/max tuple."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)
        min_v, max_v = client.supported_protocol_range()
        assert min_v == "1.0"
        assert max_v == "1.0"

    @pytest.mark.asyncio
    async def test_connect_and_close(self):
        """connect and close should work."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)

        await client.connect()
        await client.close()

    def test_sign_mcp_request_without_connect_fails(self):
        """sign_mcp_request should fail if not connected."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)

        with pytest.raises(TalosTransportError):
            client.sign_mcp_request({"method": "test"}, "tool", "action")

    @pytest.mark.asyncio
    async def test_sign_mcp_request_after_connect(self):
        """sign_mcp_request should work after connect."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)

        await client.connect()
        frame = client.sign_mcp_request({"method": "test"}, "tool", "action")

        assert frame.correlation_id.startswith("corr-")
        assert len(frame.signature) == 64

        await client.close()

    @pytest.mark.asyncio
    async def test_sign_and_send_mcp(self):
        """sign_and_send_mcp should return response."""
        wallet = Wallet.generate()
        client = TalosClient("wss://example.com", wallet)

        await client.connect()
        response = await client.sign_and_send_mcp({"method": "test"}, "tool", "action")

        assert response["status"] == "ok"
        assert "correlation_id" in response

        await client.close()

    def test_sign_http_request(self):
        """sign_http_request should produce strict output."""
        from unittest.mock import Mock, patch
        import base64
        import json
        
        # Mock wallet to capture signing input
        wallet = Mock()
        wallet.key_id = "key-test"
        wallet.sign.return_value = b"mock_signature_bytes"
        
        client = TalosClient("wss://example.com", wallet)
        
        # Fixed time and nonce
        with patch('time.time', return_value=1700000000), \
             patch('os.urandom', return_value=b'123456789012'):
            
            headers = client.sign_http_request(
                method="POST",
                path="/v1/chat",
                query="k=v",
                body={"foo": "bar"},
                opcode="test.op"
            )
            
            assert headers["X-Talos-Key-ID"] == "key-test"
            assert headers["X-Talos-Timestamp"] == "1700000000"
            assert headers["X-Talos-Sig-Alg"] == "ed25519"
            
            # Nonce check: base64url of b'123456789012'
            # MTIzNDU2Nzg5MDEy -> MTIzNDU2Nzg5MDEy (unpadded?)
            # b'123456789012' is 12 bytes.
            expected_nonce = base64.urlsafe_b64encode(b'123456789012').decode('ascii').rstrip('=')
            assert headers["X-Talos-Nonce"] == expected_nonce
            
            # Signature check: base64url of b"mock_signature_bytes"
            expected_sig = base64.urlsafe_b64encode(b"mock_signature_bytes").decode('ascii').rstrip('=')
            assert headers["X-Talos-Signature"] == expected_sig
            
            # Strict Signing Input Check
            wallet.sign.assert_called_once()
            args = wallet.sign.call_args[0]
            signed_bytes = args[0]
            
            # Expected items
            # body: canonical({"foo":"bar"}) -> b'{"foo":"bar"}'
            # method: POST
            # path_query: /v1/chat?k=v
            # nonce: ...
            # timestamp: 1700000000
            # opcode: test.op
            
            expected_input = (
                b'{"foo":"bar"}' + b"\n" +
                b"POST" + b"\n" +
                b"/v1/chat?k=v" + b"\n" +
                expected_nonce.encode('ascii') + b"\n" +
                b"1700000000" + b"\n" +
                b"test.op"
            )
            
            assert signed_bytes == expected_input

    def test_sign_http_request_empty_body(self):
        """sign_http_request should handle empty body correctly."""
        from unittest.mock import Mock, patch
        wallet = Mock()
        wallet.key_id = "key-test"
        wallet.sign.return_value = b"sig"
        client = TalosClient("wss://example.com", wallet)
        
        with patch('time.time', return_value=1), patch('os.urandom', return_value=b'x'*12):
             client.sign_http_request("GET", "/test", body=None)
             
             signed_bytes = wallet.sign.call_args[0][0]
             assert signed_bytes.startswith(b"\n") # Empty body -> b""



"""Framing logic for Talos SDK."""

import base64
import json

from .canonical import canonical_json_bytes


def base64url_encode(b: bytes) -> str:
    return base64.urlsafe_b64encode(b).decode("utf-8").rstrip("=")


def base64url_decode(s: str) -> bytes:
    s += "=" * ((4 - len(s) % 4) % 4)
    return base64.urlsafe_b64decode(s)


class Frame:
    def __init__(
        self, frame_type: str, payload: bytes, version: int = 1, flags: int = 0
    ):
        self.type = frame_type
        self.payload = payload
        self.version = version
        self.flags = flags

    def encode(self) -> bytes:
        frame_obj = {
            "version": self.version,
            "type": self.type,
            "flags": self.flags,
            "payload": base64url_encode(self.payload),
        }
        # Encode whole object as base64url
        json_bytes = canonical_json_bytes(frame_obj)
        return base64url_encode(json_bytes).encode("utf-8")

    @classmethod
    def decode(cls, data: bytes) -> "Frame":
        try:
            # Data is base64url encoded JSON
            json_str = base64url_decode(data.decode("utf-8")).decode("utf-8")
            obj = json.loads(json_str)

            if "type" not in obj or "payload" not in obj:
                raise ValueError("Invalid frame fields")

            return cls(
                frame_type=obj["type"],
                payload=base64url_decode(obj["payload"]),
                version=obj.get("version", 1),
                flags=obj.get("flags", 0),
            )
        except Exception as e:
            from .errors import TalosFrameInvalidError

            raise TalosFrameInvalidError(f"Frame decoding failed: {e}")



"""JSON Schema validation for Talos identities."""

import json
import os
from typing import Any
from jsonschema import validate, ValidationError, Draft201909Validator
from .errors import TalosInvalidInputError

class IdentityValidationError(TalosInvalidInputError):
    """Raised when an identity fails validation against normative schemas."""
    def __init__(self, message: str, path: str = None, validator_code: str = None, **kwargs: Any):
        super().__init__(message, **kwargs)
        self.path = path
        self.validator_code = validator_code

def _get_schema_path(schema_name: str) -> str:
    """Locate the schema file for the given rbac type."""
    # 1. Check for explicit override (dev/test)
    contracts_dir = os.environ.get("TALOS_CONTRACTS_DIR")
    if contracts_dir:
        path = os.path.join(contracts_dir, "schemas", "rbac", f"{schema_name}.schema.json")
        if os.path.exists(path):
            return path

    # 2. Try bundled assets (relative to this file)
    # Note: This assumes talos-contracts is installed and assets are properly bundled.
    # In this monorepo environment, we might need a better fallback.
    try:
        import talos_contracts
        package_path = os.path.dirname(talos_contracts.__file__)
        path = os.path.join(package_path, "assets", "schemas", "rbac", f"{schema_name}.schema.json")
        if os.path.exists(path):
            return path
    except ImportError:
        pass

    # 3. Monorepo fallback (relative to this file)
    monorepo_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../../contracts"))
    path = os.path.join(monorepo_path, "schemas", "rbac", f"{schema_name}.schema.json")
    if os.path.exists(path):
        return path

    raise IdentityValidationError(f"Schema for {schema_name} not found")

_VALIDATORS: dict[str, Draft201909Validator] = {}

def validate_identity(identity: dict[str, Any], schema_type: str) -> None:
    """
    Validate an identity object (Principal, Org, Team) against normative Draft 2020-12 schemas.
    
    Args:
        identity: The identity object to validate.
        schema_type: One of 'principal', 'org', 'team'.
        
    Raises:
        IdentityValidationError: If validation fails or schema is missing.
    """
    global _VALIDATORS
    
    if schema_type not in _VALIDATORS:
        schema_path = _get_schema_path(schema_type)
        try:
            with open(schema_path, "r") as f:
                schema = json.load(f)
            
            # We use Draft201909Validator explicitly for hardening
            Draft201909Validator.check_schema(schema)
            _VALIDATORS[schema_type] = Draft201909Validator(schema)
        except (ValidationError, Exception) as e:
            raise IdentityValidationError(f"Failed to load/compile schema {schema_type}: {str(e)}")

    try:
        validator = _VALIDATORS[schema_type]
        errors = list(validator.iter_errors(identity))
        
        if errors:
            # Format first error for the message
            err = errors[0]
            error_path = ".".join(map(str, err.path)) or "root"
            msg = f"Validation failed for {schema_type}: {err.message} at {error_path}"
            raise IdentityValidationError(msg, path=error_path, validator_code=err.validator)
            
    except (ValidationError, Exception) as e:
        if isinstance(e, IdentityValidationError):
            raise
        raise IdentityValidationError(f"Schema validation error: {str(e)}")

def validate_principal(principal: dict[str, Any]) -> None:
    validate_identity(principal, "principal")

def validate_org(org: dict[str, Any]) -> None:
    validate_identity(org, "org")

def validate_team(team: dict[str, Any]) -> None:
    validate_identity(team, "team")



from abc import ABC, abstractmethod

import pytest

from talos_sdk.container import Container


class IService(ABC):
    @abstractmethod
    def do(self) -> str: ...


class ServiceImpl(IService):
    def do(self) -> str:
        return "done"


def test_container_singleton() -> None:
    c = Container()
    impl = ServiceImpl()
    c.register(IService, impl)  # type: ignore[type-abstract]

    assert c.resolve(IService) is impl  # type: ignore[type-abstract]
    assert c.resolve(IService).do() == "done"  # type: ignore[type-abstract]


def test_container_factory() -> None:
    c = Container()
    c.register_factory(IService, lambda: ServiceImpl())  # type: ignore[type-abstract]

    s1 = c.resolve(IService)  # type: ignore[type-abstract]
    s2 = c.resolve(IService)  # type: ignore[type-abstract]
    assert s1 is not s2
    assert s1.do() == "done"


def test_resolve_missing() -> None:
    c = Container()
    with pytest.raises(ValueError, match="No registration found"):
        c.resolve(IService)  # type: ignore[type-abstract]



"""Talos SDK Errors.

Canonical error taxonomy as defined in SDK_CONTRACT.md.
"""

from typing import Any


class TalosError(Exception):
    """Base class for all Talos SDK errors."""

    def __init__(
        self,
        code: str,
        message: str,
        details: dict | None = None,
        request_id: str | None = None,
        cause: Exception | None = None,
    ):
        """Initialize a TalosError.

        Args:
            code: Error code from ERROR_TAXONOMY.md (e.g., TALOS_DENIED)
            message: Human-readable error message
            details: Additional context (optional)
            request_id: Correlation ID if available (optional)
            cause: Underlying exception if chained (optional)
        """
        super().__init__(message)
        self.code = code
        self.message = message
        self.details = details or {}
        self.request_id = request_id
        self.cause = cause

    def to_dict(self) -> dict:
        """Convert error to dictionary representation."""
        return {
            "code": self.code,
            "message": self.message,
            "details": self.details,
            "request_id": self.request_id,
            "cause": str(self.cause) if self.cause else None,
        }


# Authorization Errors
class TalosDeniedError(TalosError):
    """Authorization denied."""

    def __init__(self, message: str = "Authorization denied", **kwargs: Any) -> None:
        super().__init__("TALOS_DENIED", message, **kwargs)


class TalosInvalidCapabilityError(TalosError):
    """Capability token invalid."""

    def __init__(
        self, message: str = "Invalid capability token", **kwargs: Any
    ) -> None:
        super().__init__("TALOS_INVALID_CAPABILITY", message, **kwargs)


class TalosRevokedError(TalosError):
    """Identity or key revoked."""

    def __init__(self, message: str = "Identity revoked", **kwargs: Any) -> None:
        super().__init__("TALOS_REVOKED", message, **kwargs)


# Protocol Errors
class TalosProtocolMismatchError(TalosError):
    """Protocol version incompatible."""

    def __init__(
        self, message: str = "Protocol version mismatch", **kwargs: Any
    ) -> None:
        super().__init__("TALOS_PROTOCOL_MISMATCH", message, **kwargs)


class TalosFrameInvalidError(TalosError):
    """Wire frame decode failure."""

    def __init__(self, message: str = "Invalid frame", **kwargs: Any) -> None:
        super().__init__("TALOS_FRAME_INVALID", message, **kwargs)


# Crypto Errors
class TalosCryptoError(TalosError):
    """Cryptographic operation failed."""

    def __init__(
        self, message: str = "Cryptographic operation failed", **kwargs: Any
    ) -> None:
        super().__init__("TALOS_CRYPTO_ERROR", message, **kwargs)


class TalosInvalidInputError(TalosError):
    """Invalid input parameters."""

    def __init__(self, message: str = "Invalid input", **kwargs: Any) -> None:
        super().__init__("TALOS_INVALID_INPUT", message, **kwargs)


# Transport Errors
class TalosTransportTimeoutError(TalosError):
    """Transport operation timed out."""

    def __init__(self, message: str = "Transport timeout", **kwargs: Any) -> None:
        super().__init__("TALOS_TRANSPORT_TIMEOUT", message, **kwargs)


class TalosTransportError(TalosError):
    """Transport-level failure."""

    def __init__(self, message: str = "Transport error", **kwargs: Any) -> None:
        super().__init__("TALOS_TRANSPORT_ERROR", message, **kwargs)



import json
import os
import time
from typing import Any

from .handlers import BaseHandler, get_handler_for_file
from .reports import JUnitReport

# Tuple: (test_id, status, message, duration)
TestResult = tuple[str, str, str | None, float]


def run_conformance(vectors_path: str, report_path: str | None = None) -> bool:
    with open(vectors_path) as f:
        data = json.load(f)

    # If it is a release set (list of filenames), recurse
    if "version" in data and "vectors" in data and isinstance(data["vectors"], list):
        if len(data["vectors"]) > 0 and isinstance(data["vectors"][0], str):
            return _run_release_set(vectors_path, data)

    filename = os.path.basename(vectors_path)
    handler = get_handler_for_file(filename)
    if not handler:
        return True  # Skip files without handlers

    results, total, failures, errors, duration = _process_conformance_data(
        filename, data, handler
    )

    if report_path:
        suite_name = f"Conformance.{filename.replace('.json', '')}"
        report = JUnitReport()
        _write_report(
            report, suite_name, results, total, failures, errors, duration, report_path
        )

    _print_results(filename, total, failures, errors, duration)
    return failures == 0 and errors == 0


def _process_conformance_data(
    filename: str, data: dict[str, Any], handler: BaseHandler
) -> tuple[list[TestResult], int, int, int, float]:
    total = 0
    failures = 0
    errors = 0
    results: list[TestResult] = []
    start_time = time.time()

    if "steps" in data:
        res = _run_trace_file(filename, data, handler)
        results.append(res)
        total += 1
        if res[1] == "failure":
            failures += 1
        elif res[1] == "error":
            errors += 1
    elif "vectors" in data or "test_cases" in data:
        v_results, v_total, v_failures, v_errors = _run_multi_vector_file(data, handler)
        results.extend(v_results)
        total += v_total
        failures += v_failures
        errors += v_errors
    else:
        res = _run_single_vector_file(filename, data, handler)
        results.append(res)
        total += 1
        if res[1] == "failure":
            failures += 1
        elif res[1] == "error":
            errors += 1

    duration = time.time() - start_time
    return results, total, failures, errors, duration


def _run_release_set(vectors_path: str, data: dict[str, Any]) -> bool:
    base_dir = os.path.dirname(vectors_path)
    all_success = True
    for sub_vector in data["vectors"]:
        sub_path = os.path.join(base_dir, sub_vector)
        success = run_conformance(sub_path, None)
        if not success:
            all_success = False
    return all_success


def _run_trace_file(
    filename: str, data: dict[str, Any], handler: BaseHandler
) -> TestResult:
    t0 = time.time()
    try:
        handler.run_trace(data)
        if "expected_error" in data:
            return (
                filename,
                "failure",
                "Expected error but trace succeeded",
                time.time() - t0,
            )
        return (filename, "passed", None, time.time() - t0)
    except AssertionError as e:
        return (filename, "failure", str(e), time.time() - t0)
    except Exception as e:
        if "expected_error" in data:
            expected = data["expected_error"]
            msg = str(e)
            if (
                expected.get("message_contains")
                and expected["message_contains"] not in msg
            ):
                return (
                    filename,
                    "failure",
                    f"Error mismatch. Got: {msg}",
                    time.time() - t0,
                )
            return (filename, "passed", None, time.time() - t0)
        return (filename, "error", str(e), time.time() - t0)


def _run_multi_vector_file(
    data: dict[str, Any], handler: BaseHandler
) -> tuple[list[TestResult], int, int, int]:
    vectors = data.get("vectors", data.get("test_cases", []))
    results: list[TestResult] = []
    total = 0
    failures = 0
    errors = 0
    for vec in vectors:
        total += 1
        t0 = time.time()
        test_id = vec.get("test_id", vec.get("id", "unknown"))
        try:
            handler.run_vector(vec)
            results.append((test_id, "passed", None, time.time() - t0))
        except AssertionError as e:
            failures += 1
            results.append((test_id, "failure", str(e), time.time() - t0))
        except Exception as e:
            errors += 1
            results.append((test_id, "error", str(e), time.time() - t0))
    return results, total, failures, errors


def _run_single_vector_file(
    filename: str, data: dict[str, Any], handler: BaseHandler
) -> TestResult:
    t0 = time.time()
    try:
        if "expected_error" in data:
            handler.run_negative(data)
        else:
            handler.run_vector(data)
        return (data.get("test_id", filename), "passed", None, time.time() - t0)
    except AssertionError as e:
        return (data.get("test_id", filename), "failure", str(e), time.time() - t0)
    except Exception as e:
        return (data.get("test_id", filename), "error", str(e), time.time() - t0)


def _write_report(
    report: JUnitReport,
    suite_name: str,
    results: list[TestResult],
    total: int,
    failures: int,
    errors: int,
    duration: float,
    report_path: str,
) -> None:
    suite = report.add_testsuite(suite_name, total, failures, errors, duration)
    for test_id, status, message, d in results:
        case = report.add_testcase(suite, test_id, suite_name, d)
        if status == "failure":
            report.add_failure(case, message or "Unknown failure")
        elif status == "error":
            report.add_error(case, message or "Unknown error")
    report.write(report_path)
    print(f"Report written to {report_path}")


def _print_results(
    filename: str, total: int, failures: int, errors: int, duration: float
) -> None:
    print(f"Ran {total} tests in {duration:.4f}s")
    if failures or errors:
        print(f"FAILED (failures={failures}, errors={errors})")
    else:
        print("OK")
    if (failures or errors) and total > 0:
        print(f"\nFailures in {filename}:")



import os
import binascii
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from .kek import KekProvider, Envelope, ALGORITHM_AES_256_GCM

class LocalKekProvider(KekProvider):
    """
    KekProvider using a local AES key (32 bytes).
    WARN: In production, use HSM/KMS.
    """
    def __init__(self, master_key_hex: str, key_id: str):
        self.key_bytes = binascii.unhexlify(master_key_hex)
        if len(self.key_bytes) != 32:
            raise ValueError("Master key must be 32 bytes (64 hex chars)")
        self.key_id = key_id
        self.aesgcm = AESGCM(self.key_bytes)

    def encrypt(self, plaintext: bytes) -> Envelope:
        iv = os.urandom(12)
        ciphertext_and_tag = self.aesgcm.encrypt(iv, plaintext, None)
        
        # Split (tag is last 16 bytes for AESGCM in cryptography lib?)
        # Wait, AESGCM.encrypt returns ciphertext + tag appended.
        ciphertext = ciphertext_and_tag[:-16]
        tag = ciphertext_and_tag[-16:]
        
        return Envelope(
            kek_id=self.key_id,
            iv=binascii.hexlify(iv).decode('ascii'),
            ciphertext=binascii.hexlify(ciphertext).decode('ascii'),
            tag=binascii.hexlify(tag).decode('ascii'),
            alg=ALGORITHM_AES_256_GCM
        )

    def decrypt(self, envelope: Envelope) -> bytes:
        if envelope.kek_id != self.key_id:
            raise ValueError(f"Key mismatch: Envelope uses {envelope.kek_id}, Provider has {self.key_id}")
        
        if envelope.alg != ALGORITHM_AES_256_GCM:
            raise ValueError(f"Unsupported algorithm: {envelope.alg}")
            
        iv = binascii.unhexlify(envelope.iv)
        tag = binascii.unhexlify(envelope.tag)
        ciphertext = binascii.unhexlify(envelope.ciphertext)
        
        # cryptography expects ciphertext + tag for decrypt
        data = ciphertext + tag
        
        return self.aesgcm.decrypt(iv, data, None)



from abc import ABC, abstractmethod


class ICryptoPort(ABC):
    """Port for cryptographic operations (signing, verification).
    Abstracts the underlying crypto implementation (e.g. Ed25519).
    """

    @abstractmethod
    def sign(self, data: bytes, key: bytes) -> bytes:
        """Sign data using a private key."""
        ...

    @abstractmethod
    def verify(self, data: bytes, sig: bytes, key: bytes) -> bool:
        """Verify a signature using a public key."""
        ...



import hashlib
import json
from typing import Any

from talos_sdk.ports.hash import IHashPort


class NativeHashAdapter(IHashPort):
    def sha256(self, data: bytes) -> bytes:
        return hashlib.sha256(data).digest()

    def canonical_hash(self, obj: Any) -> bytes:
        from talos_sdk.canonical import canonical_json_bytes
        return self.sha256(canonical_json_bytes(obj))



from talos_sdk.ports.audit_store import (
    AuditEvent,
    EventPage,
    Filters,
    IAuditStorePort,
    Stats,
    TimeWindow,
)
from talos_sdk.ports.key_value_store import IKeyValueStorePort


class InMemoryKeyValueStore(IKeyValueStorePort):
    def __init__(self) -> None:
        self.store: dict[str, bytes] = {}

    def get(self, key: str) -> bytes | None:
        return self.store.get(key)

    def put(self, key: str, value: bytes) -> None:
        self.store[key] = value

    def delete(self, key: str) -> None:
        if key in self.store:
            del self.store[key]


class InMemoryAuditStore(IAuditStorePort):
    def __init__(self) -> None:
        self.events: list[AuditEvent] = []

    def append(self, event: AuditEvent) -> None:
        self.events.append(event)

    def list(
        self,
        before: str | None = None,
        limit: int = 100,
        filters: Filters | None = None,
    ) -> EventPage:
        # Simplistic implementation: return last N events
        # Real implementation would use cursor logic
        sliced = self.events[-limit:] if self.events else []

        # We need to return an object matching EventPage Protocol
        class Page:
            def __init__(
                self, events: list[AuditEvent], next_cursor: str | None, has_more: bool
            ) -> None:
                self.events = events
                self.next_cursor = next_cursor
                self.has_more = has_more

        # For this simple implementation, has_more is always False
        return Page(sliced, None, False)

    def stats(self, window: TimeWindow) -> Stats:
        count = 0
        for e in self.events:
            if window.start <= e.timestamp <= window.end:
                count += 1

        class StatObj:
            def __init__(self, c: int) -> None:
                self.count = c

        return StatObj(count)



"""Unit tests for RatchetFrameCrypto (Phase 10.3)."""

import pytest
from unittest.mock import MagicMock, patch
import json
import hashlib

from talos_sdk.a2a.ratchet_crypto import (
    RatchetFrameCrypto,
    _b64u_decode_strict,
    NONCE_LEN,
)
from talos_sdk.crypto import b64u_encode


# === Base64url Strict Decode Tests ===


class TestB64uDecodeStrict:
    def test_valid_decode(self):
        """Valid base64url without padding decodes correctly."""
        # "hello" in base64url
        result = _b64u_decode_strict("aGVsbG8")
        assert result == b"hello"

    def test_rejects_padding(self):
        """Inputs with padding are rejected."""
        with pytest.raises(ValueError, match="must not contain padding"):
            _b64u_decode_strict("aGVsbG8=")

    def test_rejects_invalid_length(self):
        """Inputs with len % 4 == 1 are rejected (impossible base64url)."""
        # "a" has length 1, which is impossible for valid base64url
        with pytest.raises(ValueError, match="Invalid base64url length"):
            _b64u_decode_strict("a")

        # "abcde" has length 5, len % 4 == 1
        with pytest.raises(ValueError, match="Invalid base64url length"):
            _b64u_decode_strict("abcde")

    def test_valid_lengths(self):
        """Valid lengths decode correctly."""
        # len % 4 == 0 (no padding needed)
        assert _b64u_decode_strict("YWJj") == b"abc"

        # len % 4 == 2 (needs 2 padding)
        assert _b64u_decode_strict("YWI") == b"ab"

        # len % 4 == 3 (needs 1 padding)
        assert _b64u_decode_strict("YQ") == b"a"


# === RatchetFrameCrypto Tests ===


class TestRatchetFrameCrypto:
    @pytest.fixture
    def mock_session(self):
        """Create a mock Session with encrypt/decrypt."""
        session = MagicMock()

        # Mock encrypt to return a valid envelope
        def mock_encrypt(plaintext):
            nonce = b"\x00" * 12  # 12-byte nonce
            ciphertext = b"encrypted:" + plaintext  # Fake ciphertext
            envelope = {
                "header": {"dh": "AAAA", "pn": 0, "n": 0},
                "nonce": b64u_encode(nonce),
                "ciphertext": b64u_encode(ciphertext),
            }
            from talos_sdk.canonical import canonical_json_bytes

            return canonical_json_bytes(envelope)

        session.encrypt = mock_encrypt
        return session

    def test_encrypt_returns_three_values(self, mock_session):
        """encrypt() returns (header_b64u, ciphertext_b64u, ciphertext_hash)."""
        crypto = RatchetFrameCrypto(mock_session)
        result = crypto.encrypt(b"test message")

        assert len(result) == 3
        header_b64u, ciphertext_b64u, ciphertext_hash = result

        assert isinstance(header_b64u, str)
        assert isinstance(ciphertext_b64u, str)
        assert isinstance(ciphertext_hash, str)
        assert len(ciphertext_hash) == 64  # SHA-256 hex

    def test_ciphertext_hash_excludes_nonce(self, mock_session):
        """ciphertext_hash is SHA-256 of ciphertext only, not nonce."""
        crypto = RatchetFrameCrypto(mock_session)
        _, ciphertext_b64u, ciphertext_hash = crypto.encrypt(b"test")

        # Decode and split
        combined = _b64u_decode_strict(ciphertext_b64u)
        nonce = combined[:NONCE_LEN]
        ciphertext = combined[NONCE_LEN:]

        # Hash should be of ciphertext only
        expected_hash = hashlib.sha256(ciphertext).hexdigest()
        assert ciphertext_hash == expected_hash

        # Hash should NOT include nonce
        wrong_hash = hashlib.sha256(combined).hexdigest()
        assert ciphertext_hash != wrong_hash

    def test_nonce_length_is_12(self, mock_session):
        """Nonce is exactly 12 bytes (regression guard)."""
        crypto = RatchetFrameCrypto(mock_session)
        _, ciphertext_b64u, _ = crypto.encrypt(b"test")

        combined = _b64u_decode_strict(ciphertext_b64u)
        assert len(combined) >= NONCE_LEN

        nonce = combined[:NONCE_LEN]
        assert len(nonce) == 12

    def test_header_is_canonical_json(self, mock_session):
        """header_b64u is canonical JSON bytes."""
        crypto = RatchetFrameCrypto(mock_session)
        header_b64u, _, _ = crypto.encrypt(b"test")

        header_bytes = _b64u_decode_strict(header_b64u)
        header_dict = json.loads(header_bytes.decode("utf-8"))

        # Verify keys are sorted (canonical)
        assert list(header_dict.keys()) == sorted(header_dict.keys())

    def test_encrypt_rejects_unexpected_nonce_length(self):
        """encrypt() raises if Session returns wrong nonce length."""
        session = MagicMock()

        def bad_encrypt(plaintext):
            nonce = b"\x00" * 8  # Wrong: 8 bytes instead of 12
            envelope = {
                "header": {"dh": "AAAA", "pn": 0, "n": 0},
                "nonce": b64u_encode(nonce),
                "ciphertext": b64u_encode(b"ct"),
            }
            from talos_sdk.canonical import canonical_json_bytes

            return canonical_json_bytes(envelope)

        session.encrypt = bad_encrypt
        crypto = RatchetFrameCrypto(session)

        with pytest.raises(ValueError, match="Unexpected nonce length"):
            crypto.encrypt(b"test")


class TestRatchetFrameCryptoDecrypt:
    @pytest.fixture
    def mock_session_decrypt(self):
        """Create a mock Session with decrypt."""
        session = MagicMock()
        session.decrypt.return_value = b"decrypted plaintext"
        return session

    def test_decrypt_with_valid_hash(self, mock_session_decrypt):
        """decrypt() succeeds with valid ciphertext_hash."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        # Prepare valid inputs
        header_dict = {"dh": "AAAA", "n": 0, "pn": 0}
        from talos_sdk.canonical import canonical_json_bytes

        header_b64u = b64u_encode(canonical_json_bytes(header_dict))

        nonce = b"\x00" * 12
        ciphertext = b"encrypted data"
        ciphertext_b64u = b64u_encode(nonce + ciphertext)
        ciphertext_hash = hashlib.sha256(ciphertext).hexdigest()

        result = crypto.decrypt(header_b64u, ciphertext_b64u, ciphertext_hash)
        assert result == b"decrypted plaintext"

    def test_decrypt_rejects_hash_mismatch(self, mock_session_decrypt):
        """decrypt() raises if ciphertext_hash doesn't match."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        header_dict = {"dh": "AAAA", "n": 0, "pn": 0}
        from talos_sdk.canonical import canonical_json_bytes

        header_b64u = b64u_encode(canonical_json_bytes(header_dict))

        nonce = b"\x00" * 12
        ciphertext = b"encrypted data"
        ciphertext_b64u = b64u_encode(nonce + ciphertext)
        wrong_hash = "0" * 64

        with pytest.raises(ValueError, match="hash mismatch"):
            crypto.decrypt(header_b64u, ciphertext_b64u, wrong_hash)

    def test_decrypt_rejects_too_short_ciphertext(self, mock_session_decrypt):
        """decrypt() raises if combined bytes are shorter than nonce."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        header_dict = {"dh": "AAAA", "n": 0, "pn": 0}
        from talos_sdk.canonical import canonical_json_bytes

        header_b64u = b64u_encode(canonical_json_bytes(header_dict))
        ciphertext_b64u = b64u_encode(b"short")  # Only 5 bytes

        with pytest.raises(ValueError, match="too short"):
            crypto.decrypt(header_b64u, ciphertext_b64u)

    def test_decrypt_rejects_invalid_header_json(self, mock_session_decrypt):
        """decrypt() raises if header is not valid JSON."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        header_b64u = b64u_encode(b"not valid json")
        ciphertext_b64u = b64u_encode(b"\x00" * 20)

        with pytest.raises(ValueError, match="Invalid header JSON"):
            crypto.decrypt(header_b64u, ciphertext_b64u)

    def test_decrypt_rejects_missing_header_fields(self, mock_session_decrypt):
        """decrypt() raises if required header fields are missing."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        # Missing 'pn' field
        header_dict = {"dh": "AAAA", "n": 0}
        from talos_sdk.canonical import canonical_json_bytes

        header_b64u = b64u_encode(canonical_json_bytes(header_dict))
        ciphertext_b64u = b64u_encode(b"\x00" * 20)

        with pytest.raises(ValueError, match="Missing header fields"):
            crypto.decrypt(header_b64u, ciphertext_b64u)

    def test_decrypt_rejects_padded_b64url(self, mock_session_decrypt):
        """decrypt() rejects base64url with padding."""
        crypto = RatchetFrameCrypto(mock_session_decrypt)

        with pytest.raises(ValueError, match="must not contain padding"):
            crypto.decrypt("YWJj=", "AAAA")



import base64
import json
from typing import Any

from talos_sdk.canonical import canonical_json_bytes
from talos_sdk.crypto import KeyPair
from talos_sdk.errors import TalosError
from talos_sdk.mcp import sign_mcp_request
from talos_sdk.session import PrekeyBundle, Session, SessionManager
from talos_sdk.wallet import Wallet


def base64url_decode(s: str) -> bytes:
    # Fix padding
    s += "=" * ((4 - len(s) % 4) % 4)
    return base64.urlsafe_b64decode(s)


def base64url_encode(b: bytes) -> str:
    return base64.urlsafe_b64encode(b).decode("utf-8").rstrip("=")


class BaseHandler:
    def run_vector(self, vector: dict[str, Any]) -> None:
        raise NotImplementedError

    def run_trace(self, trace: dict[str, Any]) -> None:
        raise NotImplementedError

    def run_negative(self, vector: dict[str, Any]) -> None:
        expected_error = vector.get("expected_error")
        expected_result = vector.get("expected")

        try:
            self.run_vector(vector)
        except TalosError as e:
            if expected_error:
                self._check_expected_error(vector, e.code, e.message)
                return
            raise AssertionError(f"Unexpected error when expecting result: {e}")
        except Exception as e:
            if expected_error:
                self._check_expected_error_generic(vector, str(e))
                return
            raise e

        if expected_result and "verify" in expected_result:
            return

        if expected_error:
            raise AssertionError("Expected error but operation succeeded")

    def _check_expected_error(
        self, vector: dict[str, Any], code: str, message: str
    ) -> None:
        expected = vector.get("expected_error", {})
        if "code" in expected and code != expected["code"]:
            raise AssertionError(f"Expected error code {expected['code']}, got {code}")
        if (
            "message_contains" in expected
            and expected["message_contains"].lower() not in message.lower()
        ):
            raise AssertionError(
                f"Error message should contain '{expected['message_contains']}', got '{message}'"
            )

    def _check_expected_error_generic(
        self, vector: dict[str, Any], message: str
    ) -> None:
        expected = vector.get("expected_error", {})
        if (
            "message_contains" in expected
            and expected["message_contains"].lower() not in message.lower()
        ):
            raise AssertionError(
                f"Exception message should contain '{expected['message_contains']}', got '{message}'"
            )


class SigningVerifyHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        test_id = vector.get("test_id", vector.get("id", "unknown"))
        inputs = vector["inputs"]
        expected = vector.get("expected", {})

        if test_id.startswith("sign_"):
            self._test_sign(inputs, expected)
        elif test_id.startswith("verify_"):
            self._test_verify(inputs, expected)
        else:
            raise NotImplementedError(f"Unknown test type: {test_id}")

    def _test_sign(self, inputs: dict[str, Any], expected: dict[str, Any]) -> None:
        seed_hex = inputs.get("seed_hex")
        message = inputs.get("message_utf8", "").encode("utf-8")
        wallet = Wallet.from_seed(bytes.fromhex(seed_hex)) if seed_hex else None

        if wallet:
            if "did" in expected and wallet.to_did() != expected["did"]:
                raise AssertionError(
                    f"DID mismatch: {wallet.to_did()} != {expected['did']}"
                )

            signature = wallet.sign(message)
            if "signature_base64url" in expected:
                sig_b64 = base64url_encode(signature)
                if sig_b64 != expected["signature_base64url"]:
                    raise AssertionError(
                        f"Signature mismatch. Got {sig_b64}, expected {expected['signature_base64url']}"
                    )

    def _test_verify(self, inputs: dict[str, Any], expected: dict[str, Any]) -> None:
        message = inputs.get("message_utf8", "").encode("utf-8")
        public_key = (
            bytes.fromhex(inputs["public_key_hex"])
            if "public_key_hex" in inputs
            else None
        )
        signature = (
            base64url_decode(inputs["signature_base64url"])
            if "signature_base64url" in inputs
            else None
        )

        if public_key and signature:
            result = Wallet.verify(message, signature, public_key)
            if expected.get("verify") is not None and expected.get("verify") != result:
                raise AssertionError(
                    f"Verification mismatch: {result} != {expected.get('verify')}"
                )


class CanonicalJsonHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        inputs = vector["inputs"]
        expected = vector["expected"]

        if "unordered" in inputs:
            self._handle_unordered(inputs["unordered"], expected["canonical"])
        elif "value" in inputs:
            self._handle_value(inputs["value"], expected)
        elif "pretty_printed" in inputs:
            self._handle_pretty_printed(inputs["pretty_printed"], expected["canonical"])

    def _handle_unordered(self, data: Any, expected: str) -> None:
        res = canonical_json_bytes(data).decode("utf-8")
        if res != expected:
            raise AssertionError(f"Canonical mismatch: {res} != {expected}")

    def _handle_value(self, data: Any, expected: dict[str, Any]) -> None:
        res = canonical_json_bytes(data).decode("utf-8")
        if "canonical_number" in expected and res != expected["canonical_number"]:
            raise AssertionError(
                f"Number encoding mismatch: {res} != {expected['canonical_number']}"
            )
        if "canonical" in expected and res != expected["canonical"]:
            raise AssertionError(
                f"Value canonical mismatch: {res} != {expected['canonical']}"
            )

    def _handle_pretty_printed(self, data: str, expected: str) -> None:
        obj = json.loads(data)
        res = canonical_json_bytes(obj).decode("utf-8")
        if res != expected:
            raise AssertionError(f"Pretty printed mismatch: {res} != {expected}")


class CapabilityHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        test_id = vector.get("test_id", vector.get("id"))
        inputs = vector["inputs"]
        expected = vector.get("expected", {})

        from talos_sdk.security import Capability
        from talos_sdk.wallet import Wallet

        # 1. Create capability
        issuer_wallet = Wallet.from_seed(bytes.fromhex(inputs["issuer_seed_hex"]))
        cap = Capability.create(
            issuer_wallet=issuer_wallet,
            subject_did=inputs["subject_did"],
            scope=inputs["scope"],
            exp=inputs["exp"],
            iat=inputs.get("iat", 1704067200),
        )

        # 2. Verify
        if "verify" in expected:
            # For conformance, simulate time being exactly between iat and exp
            now = (inputs.get("iat", 1704067200) + inputs["exp"]) // 2
            result = cap.verify(issuer_wallet.public_key, now=now)
            if result != expected["verify"]:
                raise AssertionError(
                    f"Capability verification mismatch for {test_id}: {result}"
                )

        # 3. Authorize
        for key, val in expected.items():
            if key.startswith("authorize_fast_"):
                parts = key.split("_")
                tool = parts[2]
                action = "_".join(parts[3:])
                result = cap.authorize(tool, action)
                if result != val:
                    raise AssertionError(
                        f"Authorization mismatch for {tool}:{action}. Got {result}"
                    )

    def run_negative(self, vector: dict[str, Any]) -> None:
        # Specific negative tests for expiry/tampering
        # For now, minimal pass if they raise the right error code
        super().run_negative(vector)


class FrameCodecHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        inputs = vector["inputs"]
        expected = vector.get("expected", {})

        from talos_sdk.frames import Frame

        if "frame_type" in inputs and "payload_utf8" in inputs:
            # Encoding test
            payload = inputs["payload_utf8"].encode("utf-8")
            frame = Frame(inputs["frame_type"], payload)
            encoded = frame.encode().decode("utf-8")
            if encoded != expected["encoded_base64url"]:
                raise AssertionError("Frame encoding mismatch")

        if "encoded_base64url" in inputs:
            # Decoding test
            frame = Frame.decode(inputs["encoded_base64url"].encode("utf-8"))
            if "frame_type" in expected and frame.type != expected["frame_type"]:
                raise AssertionError("Decoded type mismatch")
            if "version" in expected and frame.version != expected["version"]:
                raise AssertionError("Decoded version mismatch")


class MCPSignHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        inputs = vector["inputs"]
        expected = vector.get("expected", {})

        if "signer_seed_hex" in inputs:
            wallet = Wallet.from_seed(bytes.fromhex(inputs["signer_seed_hex"]))
            frame = sign_mcp_request(
                wallet,
                inputs["request"],
                inputs["session_id"],
                inputs["correlation_id"],
                inputs["tool"],
                inputs["action"],
                inputs.get("timestamp"),
            )

            if (
                "payload_canonical" in expected
                and frame.payload.decode("utf-8") != expected["payload_canonical"]
            ):
                raise AssertionError("MCP Payload mismatch")
            if (
                "signature_length" in expected
                and len(frame.signature) != expected["signature_length"]
            ):
                raise AssertionError("MCP Signature length mismatch")

        if "actual_correlation_id" in inputs:
            # Negative case: verify fail
            # verify_mcp_response needs a frame
            pass


class RatchetHandler(BaseHandler):
    def run_trace(self, trace: dict[str, Any]) -> None:
        alice_mgr, bob_mgr = self._setup_managers(trace)
        bob_bundle = self._setup_bob_prekey(trace, bob_mgr)

        alice_session = self._init_alice_session(trace, alice_mgr, bob_bundle)
        bob_session: Session | None = None

        for step in trace["steps"]:
            bob_session = self._execute_step(step, alice_session, bob_session, bob_mgr)

    def _setup_managers(
        self, trace: dict[str, Any]
    ) -> tuple[SessionManager, SessionManager]:
        def mk_pair(priv_b64: str, pub_b64: str) -> KeyPair:
            return KeyPair(
                private_key=base64url_decode(priv_b64),
                public_key=base64url_decode(pub_b64),
                key_type="ed25519",
            )

        alice_id = mk_pair(
            trace["alice"]["identity_private"], trace["alice"]["identity_public"]
        )
        bob_id = mk_pair(
            trace["bob"]["identity_private"], trace["bob"]["identity_public"]
        )
        return SessionManager(alice_id), SessionManager(bob_id)

    def _setup_bob_prekey(
        self, trace: dict[str, Any], bob_mgr: SessionManager
    ) -> PrekeyBundle:
        bob_bundle = trace["bob"]["prekey_bundle"]
        spk_priv = base64url_decode(
            trace["bob"]["bundle_secrets"]["signed_prekey_private"]
        )
        spk_pub = base64url_decode(bob_bundle["signed_prekey"])
        bob_mgr._signed_prekey = KeyPair(
            private_key=spk_priv, public_key=spk_pub, key_type="x25519"
        )
        bob_mgr._prekey_signature = base64url_decode(bob_bundle["prekey_signature"])
        return PrekeyBundle(
            identity_key=base64url_decode(bob_bundle["identity_key"]),
            signed_prekey=base64url_decode(bob_bundle["signed_prekey"]),
            prekey_signature=base64url_decode(bob_bundle["prekey_signature"]),
            one_time_prekey=base64url_decode(bob_bundle["one_time_prekey"])
            if bob_bundle["one_time_prekey"]
            else None,
        )

    def _init_alice_session(
        self, trace: dict[str, Any], alice_mgr: SessionManager, bob_bundle: PrekeyBundle
    ) -> Session:
        import talos_sdk.session

        alice_eph_priv = base64url_decode(trace["alice"]["ephemeral_private"])
        original_gen = talos_sdk.session.generate_encryption_keypair

        def mock_gen_init() -> KeyPair:
            return KeyPair(
                private_key=alice_eph_priv,
                public_key=base64url_decode(trace["steps"][0]["header"]["dh"]),
                key_type="x25519",
            )

        talos_sdk.session.generate_encryption_keypair = mock_gen_init
        try:
            return alice_mgr.create_session_as_initiator("did:bob", bob_bundle)
        finally:
            talos_sdk.session.generate_encryption_keypair = original_gen

    def _execute_step(
        self,
        step: dict[str, Any],
        alice_session: Session,
        bob_session: Session | None,
        bob_mgr: SessionManager,
    ) -> Session | None:
        import talos_sdk.session

        original_gen = talos_sdk.session.generate_encryption_keypair

        if "ratchet_priv" in step:
            r_priv = base64url_decode(step["ratchet_priv"])

            def mock_gen_step(priv: bytes = r_priv) -> KeyPair:
                from cryptography.hazmat.primitives.asymmetric import x25519
                from cryptography.hazmat.primitives.serialization import (
                    Encoding,
                    PublicFormat,
                )

                priv_obj = x25519.X25519PrivateKey.from_private_bytes(priv)
                pub_bytes = priv_obj.public_key().public_bytes(
                    Encoding.Raw, PublicFormat.Raw
                )
                return KeyPair(
                    private_key=priv, public_key=pub_bytes, key_type="x25519"
                )

            talos_sdk.session.generate_encryption_keypair = mock_gen_step

        try:
            actor = step["actor"]
            action = step["action"]
            session = alice_session if actor == "alice" else bob_session

            if action == "encrypt":
                if session is None:
                    raise ValueError(f"Session not initialized for actor {actor}")
                pt = base64url_decode(step["plaintext"])
                session.encrypt(pt)
            elif action == "decrypt":
                msg_bytes = self._reconstruct_msg(step)
                if actor == "bob" and bob_session is None:
                    bob_session = self._init_bob_responder(msg_bytes, bob_mgr)
                    session = bob_session

                if session is None:
                    raise ValueError(f"Session not initialized for actor {actor}")

                self._verify_decryption(session, msg_bytes, step)
            return bob_session
        finally:
            talos_sdk.session.generate_encryption_keypair = original_gen

    def _reconstruct_msg(self, step_data: dict[str, Any]) -> bytes:
        if "wire_message_b64u" in step_data:
            return base64url_decode(step_data["wire_message_b64u"])
        header_bytes = base64url_decode(step_data["aad"])
        nonce = base64url_decode(step_data["nonce"])
        ct = base64url_decode(step_data["ciphertext"])
        h_len = len(header_bytes).to_bytes(2, "big")
        return h_len + header_bytes + nonce + ct

    def _init_bob_responder(self, msg_bytes: bytes, bob_mgr: SessionManager) -> Session:
        try:
            envelope = json.loads(msg_bytes)
            peer_dh = base64url_decode(envelope["header"]["dh"])
        except (json.JSONDecodeError, KeyError, UnicodeDecodeError):
            h_len = int.from_bytes(msg_bytes[:2], "big")
            h_json = json.loads(msg_bytes[2 : 2 + h_len])
            peer_dh = base64url_decode(h_json["dh"])
        return bob_mgr.create_session_as_responder("did:alice", peer_dh)

    def _verify_decryption(
        self, session: Session, msg_bytes: bytes, step: dict[str, Any]
    ) -> None:
        decrypted = session.decrypt(msg_bytes)
        expected_pt = base64url_decode(
            step.get("expected_plaintext", step.get("plaintext", ""))
        )
        if decrypted != expected_pt:
            raise AssertionError(f"Decryption mismatch at step {step['step']}")


class MicroVectorHandler(BaseHandler):
    def run_vector(self, vector: dict[str, Any]) -> None:
        test_id = vector.get("test_id", vector.get("id"))
        if test_id in [
            "header_canonical_sorting",
            "header_already_sorted",
            "header_different_values",
        ]:
            self._test_header_canonical(vector)
        elif test_id == "kdf_rk_root_ratchet" or test_id == "kdf_rk_step":
            self._test_kdf_rk(vector)
        elif test_id == "kdf_ck_symmetric_ratchet" or test_id == "kdf_ck_step":
            self._test_kdf_ck(vector)
        else:
            raise NotImplementedError(f"Unknown micro-vector test: {test_id}")

    def _test_header_canonical(self, vector: dict[str, Any]) -> None:
        from talos_sdk.session import MessageHeader

        inputs = vector["input_header"]
        header = MessageHeader(
            dh_public=base64url_decode(inputs["dh"]),
            previous_chain_length=inputs["pn"],
            message_number=inputs["n"],
        )
        canonical = header.to_bytes()
        expected = base64url_decode(vector["expected_canonical_b64u"])
        if canonical != expected:
            try:
                c_obj = json.loads(canonical.decode("utf-8"))
                e_obj = json.loads(expected.decode("utf-8"))
                if c_obj == e_obj:
                    return
            except (json.JSONDecodeError, UnicodeDecodeError):
                pass
            raise AssertionError(
                f"Canonical mismatch for {vector.get('id', 'unknown')}.\nGot: {canonical.decode('utf-8')}\nExp: {expected.decode('utf-8')}"
            )

    def _test_kdf_rk(self, vector: dict[str, Any]) -> None:
        from talos_sdk.session import _kdf_rk

        rk = base64url_decode(vector["inputs"]["rk"])
        dh_out = base64url_decode(vector["inputs"]["dh_out"])
        new_rk, new_ck = _kdf_rk(rk, dh_out)
        expected_rk = base64url_decode(vector["expected"]["new_rk"])
        expected_ck = base64url_decode(vector["expected"]["new_ck"])
        if new_rk != expected_rk:
            raise AssertionError(
                f"New RK mismatch. Got {base64url_encode(new_rk)}, exp {vector['expected']['new_rk']}"
            )
        if new_ck != expected_ck:
            raise AssertionError(
                f"New CK mismatch. Got {base64url_encode(new_ck)}, exp {vector['expected']['new_ck']}"
            )

    def _test_kdf_ck(self, vector: dict[str, Any]) -> None:
        from talos_sdk.session import _kdf_ck

        ck = base64url_decode(vector["inputs"]["ck"])
        mk, next_ck = _kdf_ck(ck)
        expected_next_ck = base64url_decode(vector["expected"]["next_ck"])
        expected_mk = base64url_decode(vector["expected"]["mk"])
        if next_ck != expected_next_ck:
            raise AssertionError(
                f"Next CK mismatch. Got {base64url_encode(next_ck)}, exp {vector['expected']['next_ck']}"
            )
        if mk != expected_mk:
            raise AssertionError(
                f"MK mismatch. Got {base64url_encode(mk)}, exp {vector['expected']['mk']}"
            )


def get_handler_for_file(filename: str) -> BaseHandler | None:
    mapping: dict[str, type[BaseHandler]] = {
        "signing_verify.json": SigningVerifyHandler,
        "canonical_json.json": CanonicalJsonHandler,
        "capability_verify.json": CapabilityHandler,
        "frame_codec.json": FrameCodecHandler,
        "mcp_sign_verify.json": MCPSignHandler,
    }
    if filename in mapping:
        return mapping[filename]()

    ratchet_files = [
        "roundtrip_basic.json",
        "out_of_order.json",
        "max_skip.json",
        "v1_1_0_roundtrip.json",
    ]
    if filename in ratchet_files:
        return RatchetHandler()

    micro_files = [
        "header_canonical_bytes.json",
        "kdf_rk_step.json",
        "kdf_ck_step.json",
    ]
    if filename in micro_files:
        return MicroVectorHandler()

    return None



"""Talos SDK MCP Security.

Implements MCP request signing as defined in SDK_CONTRACT.md.
"""

import time
from dataclasses import dataclass
from typing import Any

from .canonical import canonical_json_bytes
from .wallet import Wallet


@dataclass
class SignedFrame:
    """A signed MCP frame."""

    payload: bytes
    signature: bytes
    signer_did: str
    correlation_id: str


def sign_mcp_request(
    wallet: Wallet,
    request: dict[str, Any],
    session_id: str,
    correlation_id: str,
    tool: str,
    action: str,
    timestamp: int | None = None,
) -> SignedFrame:
    """Sign an MCP request with audit bindings.

    Args:
        wallet: Signing identity wallet
        request: MCP request object (will be canonicalized)
        session_id: Session identifier
        correlation_id: Request correlation ID
        tool: Tool name being invoked
        action: Action being performed
        timestamp: Unix epoch timestamp (defaults to current time)

    Returns:
        SignedFrame containing the payload, signature, and metadata

    Note:
        Deterministic - same inputs produce identical signature.
    """
    if timestamp is None:
        timestamp = int(time.time())

    payload = {
        "request": request,
        "session_id": session_id,
        "correlation_id": correlation_id,
        "tool": tool,
        "action": action,
        "timestamp": timestamp,
    }

    payload_bytes = canonical_json_bytes(payload)
    signature = wallet.sign(payload_bytes)

    return SignedFrame(
        payload=payload_bytes,
        signature=signature,
        signer_did=wallet.to_did(),
        correlation_id=correlation_id,
    )


def verify_mcp_response(
    frame: SignedFrame,
    expected_correlation_id: str,
    signer_public_key: bytes,
) -> bool:
    """Verify a signed MCP response.

    Args:
        frame: The signed frame to verify
        expected_correlation_id: Expected correlation ID
        signer_public_key: 32-byte public key of the expected signer

    Returns:
        True if valid, False otherwise
    """
    if frame.correlation_id != expected_correlation_id:
        return False

    return Wallet.verify(frame.payload, frame.signature, signer_public_key)



"""Talos SDK Client.

High-level TalosClient facade as defined in SDK_CONTRACT.md.
"""

from typing import Any

from .errors import TalosTransportError
from .mcp import SignedFrame, sign_mcp_request
from .wallet import Wallet

# Protocol version supported by this SDK
PROTOCOL_VERSION = "1.0"
PROTOCOL_MIN = "1.0"
PROTOCOL_MAX = "1.0"


class TalosClient:
    """High-level Talos client facade.

    Composes Identity, MCP Security, and Transport modules into
    a simple interface for common operations.

    Example:
        >>> wallet = Wallet.generate("my-agent")
        >>> client = TalosClient("wss://gateway.example.com", wallet)
        >>> await client.connect()
        >>> response = await client.sign_and_send_mcp(
        ...     {"method": "read", "params": {"path": "/data"}},
        ...     "filesystem",
        ...     "read",
        ... )
        >>> await client.close()
    """

    def __init__(self, gateway_url: str, wallet: Wallet):
        """Initialize a TalosClient.

        Args:
            gateway_url: Gateway WebSocket URL
            wallet: Identity wallet for signing
        """
        self._gateway_url = gateway_url
        self._wallet = wallet
        self._connected = False
        self._session_id: str | None = None
        self._correlation_counter = 0

    @property
    def wallet(self) -> Wallet:
        """Get the client's wallet."""
        return self._wallet

    def protocol_version(self) -> str:
        """Get the negotiated protocol version.

        Returns:
            Protocol version string (e.g., "1.0")
        """
        return PROTOCOL_VERSION

    def supported_protocol_range(self) -> tuple[str, str]:
        """Get the SDK's supported protocol range.

        Returns:
            Tuple of (min_version, max_version)
        """
        return (PROTOCOL_MIN, PROTOCOL_MAX)

    async def connect(self) -> None:
        """Connect to the gateway.

        Raises:
            TalosTransportError: If connection fails
            TalosProtocolMismatchError: If protocol version incompatible
        """
        # TODO: Implement actual WebSocket connection
        # For now, mark as connected for facade completeness
        self._connected = True
        self._session_id = f"session-{id(self)}"

    async def close(self) -> None:
        """Gracefully close the connection."""
        self._connected = False
        self._session_id = None

    def _next_correlation_id(self) -> str:
        """Generate the next correlation ID."""
        self._correlation_counter += 1
        return f"corr-{self._correlation_counter}"

    def sign_mcp_request(
        self,
        request: dict[str, Any],
        tool: str,
        action: str,
    ) -> SignedFrame:
        """Sign an MCP request.

        Args:
            request: MCP request object
            tool: Tool name
            action: Action name

        Returns:
            SignedFrame ready to send

        Raises:
            TalosTransportError: If not connected
        """
        if not self._session_id:
            raise TalosTransportError("Not connected - call connect() first")

        correlation_id = self._next_correlation_id()
        return sign_mcp_request(
            self._wallet,
            request,
            self._session_id,
            correlation_id,
            tool,
            action,
        )


    async def sign_and_send_mcp(
        self,
        request: dict[str, Any],
        tool: str,
        action: str,
    ) -> dict[str, Any]:
        """Sign and send an MCP request, returning the response.

        Args:
            request: MCP request object
            tool: Tool name
            action: Action name

        Returns:
            Response from the gateway

        Raises:
            TalosTransportError: If not connected or send fails
        """
        if not self._connected:
            raise TalosTransportError("Not connected - call connect() first")

        frame = self.sign_mcp_request(request, tool, action)

        # TODO: Implement actual send/receive over WebSocket
        # For now, return a placeholder response
        return {
            "status": "ok",
            "correlation_id": frame.correlation_id,
        }

    def sign_http_request(
        self,
        method: str,
        path: str,
        query: str = "",
        body: dict[str, Any] | None = None,
        opcode: str = "http.request"
    ) -> dict[str, str]:
        """Sign an HTTP request for Phase 3 Attestation.

        Args:
            method: HTTP Method (e.g. POST)
            path: Raw path (e.g. /v1/chat)
            query: Raw query string (e.g. k=v&a=b)
            body: Request body dict (or None)
            opcode: Operation code (default: http.request)

        Returns:
            Dict of headers to add to the request:
            - X-Talos-Key-ID
            - X-Talos-Timestamp
            - X-Talos-Nonce
            - X-Talos-Signature
            - X-Talos-Sig-Alg
            - X-Talos-Sig-Version
        """
        import time
        import base64
        import json
        import os
        from .canonical import canonical_json_bytes
        
        # 1. Prepare inputs
        timestamp = int(time.time())
        nonce = base64.urlsafe_b64encode(os.urandom(12)).decode('ascii').rstrip('=')
        
        if body is None:
            body_bytes = b""
        else:
            body_bytes = canonical_json_bytes(body)
            
        method_ascii = method.upper().encode('ascii')
        
        # Path+Query: raw string exactly as sent
        # Expect caller to provide raw path and raw query string
        full_path = path + (f"?{query}" if query else "")
        path_query_ascii = full_path.encode('ascii')
        
        nonce_ascii = nonce.encode('ascii')
        ts_ascii = str(timestamp).encode('ascii')
        opcode_ascii = opcode.encode('ascii')
        
        # 2. Construct Signing Input (Strict Byte-Level)
        signing_input = (
            body_bytes + b"\n" +
            method_ascii + b"\n" +
            path_query_ascii + b"\n" +
            nonce_ascii + b"\n" +
            ts_ascii + b"\n" +
            opcode_ascii
        )
        
        # 3. Sign
        sig_bytes = self._wallet.sign(signing_input)
        sig_b64 = base64.urlsafe_b64encode(sig_bytes).decode('ascii').rstrip('=')
        
        return {
            "X-Talos-Key-ID": self._wallet.key_id,
            "X-Talos-Timestamp": str(timestamp),
            "X-Talos-Nonce": nonce,
            "X-Talos-Signature": sig_b64,
            "X-Talos-Sig-Alg": "ed25519",
            "X-Talos-Sig-Version": "v1"
        }
